from __future__ import annotations

import copy
import gzip
import logging
import os
import pickle
import tempfile
import uuid
from collections import defaultdict
from types import SimpleNamespace
from typing import Callable, Dict, List, Mapping, Optional, Sequence, Union

import cytoolz as tz
import datasets
import numpy as np
import pyarrow as pa
import torch
from datasets import DatasetInfo, Features, NamedSplit
from joblib import Parallel, delayed
from torch.utils.data._utils.collate import default_collate
from tqdm.auto import tqdm

from robustnessgym.core.dataformats.abstract import AbstractDataset
from robustnessgym.core.tools import convert_to_batch_fn

logger = logging.getLogger(__name__)

Example = Dict
Batch = Dict[str, List]


class RGImage:
    """This class acts as an interface to allow the user to manipulate the
    images without actually loading them into memory."""

    def __init__(self, filepath: str, transform: callable = None):
        self.filepath = filepath
        self.name = os.path.split(filepath)[-1]

        # Cache the transforms applied on the image when VisionDataset.update
        # gets called
        self.transform = transform

    def display(self):
        pass

    def load(self):
        import torchvision.datasets.folder as folder

        image = torch.from_numpy(np.array(folder.default_loader(self.filepath)))

        if self.transform is not None:
            image = self.transform(image)

        return image

    def __getitem__(self, idx):
        image = self.load()
        return image[idx]

    def __str__(self):
        return self.filepath

    def __repr__(self):
        return "Image(%s)" % self.name

    def __eq__(self, other):
        filepath_eq = self.filepath == other.filepath
        transform_eq = self.transform == other.transform
        return filepath_eq and transform_eq


# TODO(sabri): discuss merging RGImageRow and RGImageBatch into one class
class RGImageRow(dict):
    """Instances of this class are returned from
    `VisionDataset.__getitem__(index)` when a single row is indexed (i.e.
    `index` is an `int` or `np.int`).

    These instances should behave just as regular dictionaries do and
    should be largely indistinguishable to the end user with the only
    difference being that image columns are loaded lazily.
    """

    @classmethod
    def from_dict(
        cls,
        row: Dict[str, List],
        img_columns: Union[str, List[str]],
    ):
        row = cls(**row)
        if isinstance(img_columns, str):
            img_columns = [img_columns]
        row._img_columns = img_columns

        return row

    def __getitem__(self, key: str):
        val = super(RGImageRow, self).__getitem__(key)
        if key in self._img_columns:
            return val.load()

        return super(RGImageRow, self).__getitem__(key)


class RGImageBatch(dict):
    """Instances of this class are returned from
    `VisionDataset.__getitem__(index)` when multiple rows are indexed (i.e.
    `index` is a `slice`, `np.ndarray`, etc.).

    To the end user, these instances should be largely indistinguishable
    from a batch represented by dictionary (like those returned by
    `InMemoryDataset`) with the main difference being being that image
    columns are loaded lazily and collated (i.e. stacked) into one
    tensor only when those columns are accessed.
    """

    @classmethod
    def from_dict(
        cls,
        batch: Dict[str, List],
        img_columns: Union[str, List[str]],
        collate_fn: callable,
    ):
        batch = cls(**batch)
        batch.collate_fn = collate_fn
        if isinstance(img_columns, str):
            img_columns = [img_columns]
        batch._img_columns = img_columns

        return batch

    def __getitem__(self, key: str):
        val = super(RGImageBatch, self).__getitem__(key)

        # in `VisionDataset.map`, we use a torch DataLoader to load the images and then
        # set the value in the batch to the loaded tensor. In this case
        # batch[img_key] will already be a collated tensor, hence the check for
        # `torch.is_tensor`
        if key in self._img_columns and not torch.is_tensor(val):
            return self.collate_fn([img.load() for img in val])

        return super(RGImageBatch, self).__getitem__(key)


def save_image(image, filename):
    """Save 'image' to file 'filename' and return an RGImage object."""
    if isinstance(image, torch.Tensor):
        image = image.numpy()
    from PIL import Image as im

    image = im.fromarray(image.astype(np.uint8))
    if image.mode != "RGB":
        image = image.convert("RGB")
    image.save(filename)

    return RGImage(filename)


class VisionDataset(AbstractDataset):
    """Class for vision datasets that are to be stored in memory."""

    def __init__(
        self,
        *args,
        column_names: List[str] = None,
        info: DatasetInfo = None,
        split: Optional[NamedSplit] = None,
        img_columns: Union[str, List[str]] = ["image_file"],
        transform: Union[callable, Mapping[str, callable]] = None,
    ):

        # Data is a dictionary of lists
        self._data = {}

        if isinstance(img_columns, str):
            img_columns = [img_columns]
        self._img_columns = img_columns

        self._img_key_to_transform = defaultdict(lambda: None)
        if isinstance(transform, Mapping):
            if (set(transform.keys()) - self._img_columns) != 0:
                raise ValueError(
                    "Mapping passed to `transforms` includes keys not in `img_columns`."
                )
            self._img_key_to_transform.update(transform)
        elif transform is not None:
            # if single `callable` is passed, use same transform for all image columns
            self._img_key_to_transform.update({k: transform for k in self._img_columns})

        self.collate_fn = default_collate

        # Internal state variables for the update function
        self._updating_images = False
        self._adding_images = False
        self._callstack = []

        # Single argument
        if len(args) == 1:
            assert column_names is None, "Don't pass in column_names."
            # The data is passed in
            data = args[0]

            # TODO: Discuss whether this is a worthy refactor
            # This replaces the commented elif block below
            if isinstance(data, list) and len(data):
                # Transpose the list of dicts to a dict of lists i.e. a batch
                data = tz.merge_with(list, *data)

            # `data` is a dictionary
            if isinstance(data, dict) and len(data):
                # Assert all columns are the same length
                self._assert_columns_all_equal_length(data)
                mask = [key in data for key in self._img_columns]
                if not all(mask):
                    idx = mask.index(False)
                    raise KeyError(
                        "Key with paths to images not found: %s"
                        % self._img_columns[idx]
                    )
                for key in self._img_columns:
                    self._paths_to_Images(data, key)
                self._data = data

            # `data` is a list
            # elif isinstance(data, list) and len(data):
            #     # Transpose the list of dicts to a dict of lists i.e. a batch
            #     data = tz.merge_with(list, *data)
            #     # Assert all columns are the same length
            #     self._assert_columns_all_equal_length(data)
            #     if filepath_key not in data:
            #         raise KeyError(
            #             "Key with paths to images not found: %s" % filepath_key
            #         )
            #     self._paths_to_Images(data, filepath_key)
            #     self._data = data

            # `data` is a datasets.Dataset
            elif isinstance(data, datasets.Dataset):
                self._data = data[:]
                info, split = data.info, data.split

        # No argument
        elif len(args) == 0:

            # Use column_names to setup the data dictionary
            if column_names:
                self._data = {k: [] for k in column_names}

        else:
            raise NotImplementedError(
                "Currently only one table is supported when creating a VisionDataSet"
            )

        # Setup the DatasetInfo
        info = info.copy() if info is not None else DatasetInfo()
        AbstractDataset.__init__(self, info=info, split=split)

        # Create attributes for all columns and visible columns
        self.all_columns = list(self._data.keys())
        self.visible_columns = None

        # Create attributes for visible rows
        self.visible_rows = None

        # Initialization
        self._initialize_state()

        logger.info(
            f"Created `VisionDataset` with {len(self)} rows and "
            f"{len(self.column_names)} columns."
        )

    def _paths_to_Images(self, data, key):
        """Convert a list of paths to images data[key] into a list of RGImage
        instances."""
        if isinstance(data[key][0], RGImage):
            return  # Can happen when we're copying a dataset
        data[key] = [
            RGImage(i, transform=self._img_key_to_transform[key]) for i in data[key]
        ]

    def _set_features(self):
        """Set the features of the dataset."""
        with self.format():
            d = {
                k: [""] if (k in self._img_columns or torch.torch.is_tensor(v)) else [v]
                for k, v in self[0].items()
            }
            self.info.features = Features.from_arrow_schema(
                pa.Table.from_pydict(
                    d,
                ).schema
            )

    def _materialize(self):
        # Materialize data, instead of using a reference to an ancestor Dataset
        self._data = {k: self[k] for k in self._data}

        # Reset visible_rows
        self.set_visible_rows(None)

    def add_column(self, column: str, values: List, overwrite=False) -> None:
        """Add a column to the dataset."""

        assert (
            column not in self.all_columns
        ) or overwrite, (
            f"Column `{column}` already exists, set `overwrite=True` to overwrite."
        )
        assert len(values) == len(self), (
            f"`add_column` failed. "
            f"Values length {len(values)} != dataset length {len(self)}."
        )

        if self.visible_rows is not None:
            # Materialize the data
            self._materialize()

        # Add the column
        self._data[column] = list(values)
        self.all_columns.append(column)
        self.visible_columns.append(column)

        # Set features
        self._set_features()

        logger.info(f"Added column `{column}` with length `{len(values)}`.")

    def remove_column(self, column: str) -> None:
        """Remove a column from the dataset."""
        assert column in self.all_columns, f"Column `{column}` does not exist."

        # Remove the column
        del self._data[column]
        self.all_columns = [col for col in self.all_columns if col != column]
        self.visible_columns = [col for col in self.visible_columns if col != column]

        # Set features
        self._set_features()

        logger.info(f"Removed column `{column}`.")

    def select_columns(self, columns: List[str]) -> Batch:
        """Select a subset of columns."""
        for col in columns:
            assert col in self._data
        return tz.keyfilter(lambda k: k in columns, self._data)

    def _append_to_empty_dataset(self, example_or_batch: Union[Example, Batch]) -> None:
        """Append a batch of data to the dataset when it's empty."""
        # Convert to batch
        batch = self._example_or_batch_to_batch(example_or_batch)

        # TODO(karan): what other data properties need to be in sync here
        self.all_columns = list(batch.keys())
        self.visible_columns = list(batch.keys())

        # Dataset is empty: create the columns and append the batch
        self._data = {k: [] for k in self.column_names}
        for k in self.column_names:
            self._data[k].extend(batch[k])

    def append(
        self,
        example_or_batch: Union[Example, Batch],
    ) -> None:
        """Append a batch of data to the dataset.

        `batch` must have the same columns as the dataset (regardless of
        what columns are visible).
        """
        if not self.column_names:
            return self._append_to_empty_dataset(example_or_batch)

        # Check that example_or_batch has the same format as the dataset
        # TODO(karan): require matching on nested features?
        columns = list(example_or_batch.keys())
        assert set(columns) == set(
            self.column_names
        ), f"Mismatched columns\nbatch: {columns}\ndataset: {self.column_names}"

        # Convert to a batch
        batch = self._example_or_batch_to_batch(example_or_batch)

        # Append to the dataset
        for k in self.column_names:
            if k in self._img_columns:
                batch[k] = list(map(RGImage, batch[k]))
            self._data[k].extend(batch[k])

    def _remap_index(self, index):
        if isinstance(index, int):
            return self.visible_rows[index].item()
        elif isinstance(index, slice):
            return self.visible_rows[index].tolist()
        elif isinstance(index, str):
            return index
        elif (isinstance(index, tuple) or isinstance(index, list)) and len(index):
            return self.visible_rows[index].tolist()
        elif isinstance(index, np.ndarray) and len(index.shape) == 1:
            return self.visible_rows[index].tolist()
        else:
            raise TypeError("Invalid argument type: {}".format(type(index)))

    def __getitem__(self, index):
        if self.visible_rows is not None:
            # Remap the index if only some rows are visible
            index = self._remap_index(index)
        if isinstance(index, str):
            # str index => column selection
            if index in self.column_names:
                if self.visible_rows is not None:
                    return [self._data[index][i] for i in self.visible_rows]
                return self._data[index]
            raise AttributeError(f"Column {index} does not exist.")

        if isinstance(index, int) or isinstance(index, np.int):
            return RGImageRow.from_dict(
                {k: self._data[k][index] for k in self.visible_columns},
                img_columns=self._img_columns,
            )

        # indices that return batches
        if isinstance(index, slice):
            # int or slice index => standard list slicing
            batch = {k: self._data[k][index] for k in self.visible_columns}
        elif (isinstance(index, tuple) or isinstance(index, list)) and len(index):
            batch = {k: [self._data[k][i] for i in index] for k in self.visible_columns}
        elif isinstance(index, np.ndarray) and len(index.shape) == 1:
            batch = {
                k: [self._data[k][int(i)] for i in index] for k in self.visible_columns
            }
        else:
            raise TypeError("Invalid argument type: {}".format(type(index)))

        return RGImageBatch.from_dict(
            batch, img_columns=self._img_columns, collate_fn=self.collate_fn
        )

    def _inspect_update_function(
        self,
        function: Callable,
        with_indices: bool = False,
        batched: bool = False,
    ) -> SimpleNamespace:
        """Load the images before calling _inspect_function, and check if new
        image columns are being added."""

        properties = self._inspect_function(function, with_indices, batched)

        # Check if new columns are added
        if batched:
            if with_indices:
                output = function(self[:2], range(2))
            else:
                output = function(self[:2])

        else:
            if with_indices:
                output = function(self[0], 0)
            else:
                output = function(self[0])
        new_columns = set(output.keys()).difference(set(self.all_columns))

        # Check if any of those new columns is an image column
        new_img_columns = []
        for key in new_columns:
            val = output[key]
            if isinstance(val, torch.Tensor) and len(val.shape) >= 2:
                new_img_columns.append(key)

        properties.new_image_columns = new_img_columns

        return properties

    def update(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        # input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        remove_columns: Optional[List[str]] = None,
        cache_dir: str = None,
        **kwargs,
    ) -> Optional[VisionDataset]:
        """Update the columns of the dataset."""
        # TODO(karan): make this fn go faster
        # most of the time is spent on the merge, speed it up further

        # Sanity check when updating the images
        self._callstack.append("update")

        # Return if the function is None
        if function is None:
            logger.info("`function` None, returning None.")
            return self

        # Return if `self` has no examples
        if not len(self):
            logger.info("Dataset empty, returning None.")
            return self

        # Get some information about the function
        function_properties = self._inspect_update_function(
            function, with_indices, batched
        )
        assert (
            function_properties.dict_output
        ), f"`function` {function} must return dict."

        if not batched:
            # Convert to a batch function
            function = convert_to_batch_fn(function, with_indices=with_indices)
            logger.info(f"Converting `function` {function} to batched function.")

        updated_columns = function_properties.existing_columns_updated
        changed_images = [key in self._img_columns for key in updated_columns]
        new_image_columns = function_properties.new_image_columns

        # Set the internal state for the map function
        self._updating_images = any(changed_images)
        self._adding_images = any(new_image_columns)
        if self._updating_images or self._adding_images:
            # Set the cache directory where the modified images will be stored
            if not cache_dir:
                cache_dir = tempfile.gettempdir()
                logger.warning(
                    "Modifying the images without setting a cache directory.\n"
                    "Consider setting it if your dataset is very large.\n"
                    "The default image cache location is: {}".format(cache_dir)
                )

            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)

            cache_dir = os.path.join(cache_dir, uuid.uuid4().hex)
            os.mkdir(cache_dir)

        # Update always returns a new dataset
        logger.info("Running update, a new dataset will be returned.")
        if self.visible_rows is not None:
            # Run .map() to get updated batches and pass them into a new dataset
            new_dataset = VisionDataset(
                self.map(
                    (
                        lambda batch, indices: self._merge_batch_and_output(
                            batch, function(batch, indices)
                        )
                    )
                    if with_indices
                    else (
                        lambda batch: self._merge_batch_and_output(
                            batch, function(batch)
                        )
                    ),
                    with_indices=with_indices,
                    batched=True,
                    batch_size=batch_size,
                    cache_dir=cache_dir,
                ),
                img_columns=self._img_columns,
            )
        else:
            if function_properties.updates_existing_column:
                # Copy the ._data dict with a reference to the actual columns
                new_dataset = self.copy()

                # Calculate the values for the updated columns using a .map()
                output = self.map(
                    (
                        lambda batch, indices:
                        # Only merge columns that get updated
                        self._merge_batch_and_output(
                            {
                                k: v
                                for k, v in batch.items()
                                if k in function_properties.existing_columns_updated
                            },
                            function(batch, indices),
                        )
                    )
                    if with_indices
                    else (
                        lambda batch:
                        # Only merge columns that get updated
                        self._merge_batch_and_output(
                            {
                                k: v
                                for k, v in batch.items()
                                if k in function_properties.existing_columns_updated
                            },
                            function(batch),
                        )
                    ),
                    with_indices=with_indices,
                    batched=True,
                    batch_size=batch_size,
                    cache_dir=cache_dir,
                    new_image_columns=new_image_columns,
                )

                # If new image columns were added, update that information
                if self._adding_images:
                    new_dataset._img_columns.extend(new_image_columns)

                # Add new columns / overwrite existing columns for the update
                for col, vals in output.items():
                    if isinstance(vals[0], torch.Tensor) and vals[
                        0
                    ].shape == torch.Size([]):
                        # Scalar tensor. Convert to Python.
                        new_vals = []
                        for val in vals:
                            new_vals.append(val.item())
                        vals = new_vals
                    new_dataset.add_column(col, vals, overwrite=True)
            else:
                # Copy the ._data dict with a reference to the actual columns
                new_dataset = self.copy()

                # Calculate the values for the new columns using a .map()
                output = new_dataset.map(
                    function=function,
                    with_indices=with_indices,
                    batched=True,
                    batch_size=batch_size,
                    cache_dir=cache_dir,
                    new_image_columns=new_image_columns,
                )

                # If new image columns were added, update that information
                if self._adding_images:
                    new_dataset._img_columns.extend(new_image_columns)

                # Add new columns for the update
                for col, vals in output.items():
                    if isinstance(vals[0], torch.Tensor) and vals[
                        0
                    ].shape == torch.Size([]):
                        # Scalar tensor. Convert to Python.
                        new_vals = []
                        for val in vals:
                            new_vals.append(val.item())
                        vals = new_vals
                    new_dataset.add_column(col, vals)

        # Remove columns
        if remove_columns:
            for col in remove_columns:
                new_dataset.remove_column(col)
            logger.info(f"Removed columns {remove_columns}.")
        # Reset the format
        # if input_columns:
        #     self.set_format(previous_format)

        # Remember to reset the internal state
        self._updating_images = False
        self._adding_images = False
        # And remove this call from the callstack
        self._callstack.pop()

        # If the new dataset is a copy we also need to reset it
        new_dataset._updating_images = False
        new_dataset._adding_images = False
        new_dataset._callstack.pop()

        return new_dataset

    def map(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        drop_last_batch: bool = False,
        num_proc: Optional[int] = 64,
        **kwargs,
    ) -> Optional[Union[Dict, List]]:
        """Apply a map over the dataset."""

        # Just return if the function is None
        if function is None:
            logger.info("`function` None, returning None.")
            return None

        # Ensure that num_proc is not None
        if num_proc is None:
            num_proc = 64

        # Return if `self` has no examples
        if not len(self):
            logger.info("Dataset empty, returning None.")
            return None

        if isinstance(input_columns, str):
            input_columns = [input_columns]

        # Set the format
        previous_format = self.visible_columns
        if input_columns:
            self.set_format(input_columns)

        if not batched:
            # Convert to a batch function
            function = convert_to_batch_fn(function, with_indices=with_indices)
            logger.info(f"Converting `function` {function} to a batched function.")

        # Check if any of the columns is an image column
        if not input_columns:
            input_columns = self.visible_columns
        image_loaders = {}
        for key in input_columns:
            if key in self._img_columns:
                # Load the images
                images = self.to_dataloader(
                    columns=[key],
                    batch_size=batch_size,
                    shuffle=False,
                    drop_last=False,
                    num_workers=num_proc,
                    collate_fn=self.collate_fn,
                )
                images = iter(images)
                image_loaders[key] = images

        # If we are updating, prepare image savers and perform sanity checks
        if self._updating_images or self._adding_images:
            assert "update" in self._callstack, (
                "_updating_images and _adding_images can only be set by "
                "VisionDataset.update"
            )
            assert "cache_dir" in kwargs, "No cache directory specified"
            cache_dir = kwargs["cache_dir"]
        if self._adding_images:
            assert "new_image_columns" in kwargs, "New image column names not specified"
            new_image_columns = kwargs["new_image_columns"]

        # Run the map
        logger.info("Running `map`, the dataset will be left unchanged.")
        outputs = None
        for i, batch in tqdm(
            enumerate(self.batch(batch_size, drop_last_batch)),
            total=(len(self) // batch_size)
            + int(not drop_last_batch and len(self) % batch_size != 0),
        ):
            for key in image_loaders:
                batch[key] = next(image_loaders[key])

            # Run `function` on the batch
            output = (
                function(
                    batch,
                    range(i * batch_size, min(len(self), (i + 1) * batch_size)),
                )
                if with_indices
                else function(batch)
            )

            # Save the modified images
            if self._updating_images:
                for key in image_loaders:
                    images = output[key]

                    # Save the images in parallel
                    rgimages = Parallel(n_jobs=num_proc)(
                        delayed(save_image)(
                            images[idx],
                            os.path.join(
                                cache_dir,
                                "{0}{1}.png".format(key, i * batch_size + idx),
                            ),
                        )
                        for idx in range(len(images))
                    )

                    output[key] = rgimages

            if self._adding_images:
                for key in new_image_columns:
                    images = output[key]

                    # Save the images in parallel
                    rgimages = Parallel(n_jobs=num_proc)(
                        delayed(save_image)(
                            images[idx],
                            os.path.join(
                                cache_dir,
                                "{0}{1}.png".format(key, i * batch_size + idx),
                            ),
                        )
                        for idx in range(len(images))
                    )

                    output[key] = rgimages

            if i == 0:
                # Create an empty dict or list for the outputs
                outputs = defaultdict(list) if isinstance(output, Mapping) else []

            # Append the output
            if output is not None:
                if isinstance(output, Mapping):
                    for k in output.keys():
                        outputs[k].extend(output[k])
                else:
                    outputs.extend(output)

        # Reset the format
        if input_columns:
            self.set_format(previous_format)

        if not len(outputs):
            return None
        elif isinstance(outputs, dict):
            return dict(outputs)
        return outputs

    def filter(
        self,
        function: Optional[Callable] = None,
        with_indices=False,
        input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        drop_last_batch: bool = False,
        num_proc: Optional[int] = 64,
        **kwargs,
    ) -> Optional[VisionDataset]:
        """Apply a filter over the dataset."""
        # Just return if the function is None
        if function is None:
            logger.info("`function` None, returning None.")
            return None

        # Return if `self` has no examples
        if not len(self):
            logger.info("Dataset empty, returning None.")
            return None

        # Get some information about the function
        function_properties = self._inspect_function(
            function,
            with_indices,
            batched=batched,
        )
        assert function_properties.bool_output, "function must return boolean."

        # Map to get the boolean outputs and indices
        logger.info("Running `filter`, a new dataset will be returned.")
        outputs = self.map(
            function=function,
            with_indices=with_indices,
            input_columns=input_columns,
            batched=batched,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            num_proc=num_proc,
        )
        indices = np.where(outputs)[0]

        # Reset the format to set visible columns for the filter
        with self.format():
            # Filter returns a new dataset
            new_dataset = self.copy()
            new_dataset.set_visible_rows(indices)

        return new_dataset

    def to_dataloader(
        self,
        columns: Sequence[str],
        column_to_transform: Optional[Mapping[str, Callable]] = None,
        **kwargs,
    ) -> torch.utils.data.DataLoader:
        """Get a PyTorch dataloader that iterates over a subset of the columns
        (specified by `keys`) in the dataset. This is handy when using the dataset with
        training or evaluation loops outside of robustnessgym.  For example:
        ```
        dataset = Dataset(...)
        for img, target in dataset.to_dataloader(
            keys=["img_path", "label"],
            batch_size=16,
            num_workers=12
        ):
            out = model(img)
            loss = loss(out, target)
            ...
        ```

        Args:
            columns (Sequence[str]): A subset of the columns in the vision dataset.
                Specifies the columns to load. The dataloader will return values in same
                 order as `columns` here.
            column_to_transform (Optional[Mapping[str, Callable]], optional): A mapping
                from zero or more `keys` to callable transforms to be applied by the
                dataloader. Defaults to None, in which case no transforms are applied.
                Example: `column_to_transform={"img_path": transforms.Resize((16,16))}`.
                Note: these transforms will be applied after the transforms specified
                via the `transform` argument to `VisionDataset`.

        Returns:
            torch.utils.data.DataLoader: dataloader that iterates over dataset
        """
        img_folder = TorchDataset(
            self, columns=columns, column_to_transform=column_to_transform
        )
        return torch.utils.data.DataLoader(img_folder, **kwargs)

    def copy(self, deepcopy=False):
        """Return a copy of the dataset."""
        if deepcopy:
            return copy.deepcopy(self)
        else:
            dataset = VisionDataset()
            dataset.__dict__ = {k: copy.copy(v) for k, v in self.__dict__.items()}
            return dataset

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return {
            "_data",
            "all_columns",
            "visible_rows",
            "_info",
            "_split",
            "_img_columns",
            "_updating_images",
            "_adding_images",
            "_callstack",
        }

    @classmethod
    def _assert_state_keys(cls, state: Dict) -> None:
        """Assert that a state contains all required keys."""
        assert (
            set(state.keys()) == cls._state_keys()
        ), f"State must contain all state keys: {cls._state_keys()}."

    def __getstate__(self) -> Dict:
        """Get the internal state of the dataset."""
        state = {key: getattr(self, key) for key in self._state_keys()}
        self._assert_state_keys(state)
        return state

    def __setstate__(self, state: Dict) -> None:
        """Set the internal state of the dataset."""
        if not isinstance(state, dict):
            raise ValueError(
                f"`state` must be a dictionary containing " f"{self._state_keys()}."
            )

        self._assert_state_keys(state)

        for key in self._state_keys():
            setattr(self, key, state[key])

        # Do some initialization
        self._initialize_state()

    @classmethod
    def load_from_disk(cls, path: str) -> VisionDataset:
        """Load the in-memory dataset from disk."""

        with gzip.open(os.path.join(path, "data.gz")) as f:
            dataset = pickle.load(f)
        # # Empty state dict
        # state = {}
        #
        # # Load the data
        # with gzip.open(os.path.join(path, "data.gz")) as f:
        #     state['_data'] = pickle.load(f)
        #
        # # Load the metadata
        # metadata = json.load(
        #     open(os.path.join(path, "metadata.json"))
        # )
        #
        # # Merge the metadata into the state
        # state = {**state, **metadata}

        # Create an empty `VisionDataset` and set its state
        # dataset = cls()
        # dataset.__setstate__(state)

        return dataset

    def save_to_disk(self, path: str):
        """Save the in-memory dataset to disk."""
        # Create all the directories to the path
        os.makedirs(path, exist_ok=True)

        # Store the data in a compressed format
        with gzip.open(os.path.join(path, "data.gz"), "wb") as f:
            pickle.dump(self, f)

        # # Get the dataset state
        # state = self.__getstate__()
        #
        # # Store the data in a compressed format
        # with gzip.open(os.path.join(path, "data.gz"), "wb") as f:
        #     pickle.dump(state['_data'], f)
        #
        # # Store the metadata
        # json.dump(
        #     {k: v for k, v in state.items() if k != '_data'},
        #     open(os.path.join(path, "metadata.json"), 'w'),
        # )


class TorchDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        dataset: VisionDataset,
        columns: Sequence[str],
        column_to_transform: Optional[Mapping[str, Callable]] = None,
    ):
        """A torch dataset wrapper around `VisionDataset` that can be used in a
        torch dataloder.

        `VisionDataset.__getitem__` returns a dict containing `RGImage`
        objects, which is not compatible with the torch's default
        `collate_fn`. This dataset returns a subset of columns in the
        dataset specified by `keys` and calls `load` on any RGImage
        before returning to the `collate_fn`. It also supports
        specifying a different transformation function for each key
        returned.
        """
        self.columns = columns
        self.column_to_transform = (
            {} if column_to_transform is None else column_to_transform
        )
        self.dataset = dataset

    def __getitem__(self, index: int):
        row = self.dataset[index]
        vals = [
            self.column_to_transform.get(k, lambda x: x)(
                v.load() if isinstance(v, RGImage) else v
            )
            for k, v in ((key, row[key]) for key in self.columns)
        ]
        return tuple(vals) if len(vals) > 1 else vals[0]

    def __len__(self):
        return len(self.dataset)



from typing import List

import torch

try:
    from allennlp.predictors import Predictor
except ImportError:
    _allennlp_available = False
else:
    _allennlp_available = True

from robustnessgym.core.cachedops import SingleColumnCachedOperation


class AllenPredictor(SingleColumnCachedOperation):
    def __init__(self, path: str, device: str, *args, **kwargs):
        super(AllenPredictor, self).__init__(*args, **kwargs)

        # If no device is passed in, automatically use GPU if available
        if not device:
            device = "cuda" if torch.cuda.is_available() else "cpu"

        # Resolve the device
        cuda_device = -1
        if device.startswith("cuda"):
            cuda_device = 0 if ":" not in device else int(device.split(":")[-1])

        # Set up Allen's predictor
        self._predictor = Predictor.from_path(
            archive_path=path, cuda_device=cuda_device
        )

    @property
    def predictor(self):
        return self._predictor

    def single_column_apply(self, column_batch: List, *args, **kwargs) -> List:
        # Apply the predictor
        return self.predictor.predict_batch_json(
            [{"sentence": text} for text in column_batch]
        )



from typing import List

from robustnessgym import Identifier, SliceBuilder
from robustnessgym.core.constants import CURATION


class Curator(SliceBuilder):
    def __init__(
        self,
        identifiers: List[Identifier],
        apply_fn,
    ):
        super(Curator, self).__init__(
            category=CURATION,
            identifiers=identifiers,
            apply_fn=apply_fn,
        )

    def __call__(self, *args, **kwargs):
        pass

    def apply(self, *args, **kwargs):
        pass



from functools import partial
from types import SimpleNamespace

import numpy as np
import scipy.optimize
import scipy.stats
from scipy.special import logsumexp
from sklearn.metrics.pairwise import rbf_kernel


def Phi(D, edge_list: list = None):
    """Given an n x d matrix of (example, slices), calculate the potential
    matrix.

    Includes correlations modeled by the edges in the `edge_list`.
    """
    if edge_list is not None:
        pairwise_terms = (
            D[np.arange(len(D)), edge_list[:, 0][:, np.newaxis]].T
            * D[np.arange(len(D)), edge_list[:, 1][:, np.newaxis]].T
        )
        return np.concatenate([D, pairwise_terms], axis=1)
    else:
        return D


def log_partition_ratio(x, Phi_D_src, n_src):
    """Calculate the log-partition ratio in the KLIEP problem."""
    return np.log(n_src) - logsumexp(Phi_D_src.dot(x))


def mandoline(
    D_src,
    D_tgt,
    edge_list,
    sigma=None,
):
    """
    Mandoline solver.
    Args:
        D_src: (n_src x d) matrix of (example, slices) for the source distribution.
        D_tgt: (n_tgt x d) matrix of (example, slices) for the source distribution.
        edge_list: list of edge correlations between slices that should be modeled.
        sigma: optional parameter that activates RBF kernel-based KLIEP with scale
        `sigma`.

    Returns: SimpleNamespace that contains
        opt: result of scipy.optimize
        Phi_D_src: source potential matrix used in Mandoline
        Phi_D_tgt: target potential matrix used in Mandoline
        n_src: number of source samples
        n_tgt: number of target samples
        edge_list: the `edge_list` parameter passed as input

    """
    # Copy and binarize the input matrices to -1/1
    D_src, D_tgt = np.copy(D_src), np.copy(D_tgt)
    if np.min(D_src) == 0:
        D_src[D_src == 0] = -1
        D_tgt[D_tgt == 0] = -1

    # Edge list encoding dependencies between gs
    if edge_list is not None:
        edge_list = np.array(edge_list)

    # Create the potential matrices
    Phi_D_tgt, Phi_D_src = Phi(D_tgt, edge_list), Phi(D_src, edge_list)

    # Number of examples
    n_src, n_tgt = Phi_D_src.shape[0], Phi_D_tgt.shape[0]

    def f(x):
        obj = Phi_D_tgt.dot(x).sum() - n_tgt * logsumexp(Phi_D_src.dot(x))
        return -obj

    # Set the kernel
    kernel = partial(rbf_kernel, gamma=sigma)

    def llkliep_f(x):
        obj = kernel(Phi_D_tgt, x[:, np.newaxis]).sum() - n_tgt * logsumexp(
            kernel(Phi_D_src, x[:, np.newaxis])
        )
        return -obj

    # Solve
    if not sigma:
        opt = scipy.optimize.minimize(
            f, np.random.randn(Phi_D_tgt.shape[1]), method="BFGS"
        )
    else:
        opt = scipy.optimize.minimize(
            llkliep_f, np.random.randn(Phi_D_tgt.shape[1]), method="BFGS"
        )

    return SimpleNamespace(
        opt=opt,
        Phi_D_src=Phi_D_src,
        Phi_D_tgt=Phi_D_tgt,
        n_src=n_src,
        n_tgt=n_tgt,
        edge_list=edge_list,
    )


def log_density_ratio(D, solved):
    """Calculate the log density ratio for a solved Mandoline run."""
    Phi_D = Phi(D, None)
    return Phi_D.dot(solved.opt.x) + log_partition_ratio(
        solved.opt.x, solved.Phi_D_src, solved.n_src
    )


def get_k_most_unbalanced_gs(D_src, D_tgt, k):
    """Get the top k slices that shift most between source and target
    distributions.

    Uses difference in marginals between each slice.
    """
    marginal_diff = np.abs(D_src.mean(axis=0) - D_tgt.mean(axis=0))
    differences = np.sort(marginal_diff)[-k:]
    indices = np.argsort(marginal_diff)[-k:]
    return list(indices), list(differences)


def weighted_estimator(weights, empirical_mat):
    """Calculate a weighted empirical mean over a matrix.

    Calculates an unweighted mean if `weights` is None.
    """
    if weights is None:
        return np.mean(empirical_mat, axis=0)
    return np.sum(weights[:, np.newaxis] * empirical_mat, axis=0)


def estimate_performance(
    D_src,
    D_tgt,
    edge_list,
    empirical_mat_list_src,
):
    """Estimate performance on a target distribution using slices from the
    source and target data."""
    # Run the solver
    solved = mandoline(D_src, D_tgt, edge_list)

    # Compute the weights on the source dataset
    density_ratios = np.e ** log_density_ratio(solved.Phi_D_src, solved)

    # Self-normalized importance weights
    weights = density_ratios / np.sum(density_ratios)

    all_estimates = []
    for mat_src in empirical_mat_list_src:
        # Estimates is a 1-D array of estimates for each mat e.g.
        # each mat can correspond to a model's (n x 1) error matrix
        weighted_estimates = weighted_estimator(weights, mat_src)
        source_estimates = weighted_estimator(
            np.ones(solved.n_src) / solved.n_src, mat_src
        )

        all_estimates.append(
            SimpleNamespace(
                weighted=weighted_estimates,
                source=source_estimates,
            )
        )

    return SimpleNamespace(
        all_estimates=all_estimates,
        solved=solved,
        weights=weights,
    )



import itertools
import re
from typing import Callable, Collection, Dict, List, Optional

import cytoolz as tz

try:
    from ludwig.api import LudwigModel as ludwigmodel
except ImportError:
    _ludwig_available = False
    ludwig = None
else:
    _ludwig_available = True

try:
    import nltk
except ImportError:
    _nltk_available = False
    nltk = None
else:
    _nltk_available = True
import torch
from transformers import (
    AutoModel,
    AutoModelForSeq2SeqLM,
    AutoModelForSequenceClassification,
    AutoTokenizer,
)

from robustnessgym.core.dataset import Dataset
from robustnessgym.core.metrics import compute_metric
from robustnessgym.tasks.task import Task


class Model:
    def __init__(
        self,
        identifier: str,
        task: Task,
        model=None,
        evaluation_fn=None,
        device: str = None,
        is_classifier: bool = None,
    ):

        # TODO(karan): improve this wrapper around models
        # TODO(karan): add some human-readble identifier to this as optional
        self.identifier = identifier
        self.task = task
        self.model = model

        if evaluation_fn is not None:
            self.evaluate = evaluation_fn

        if self.task is None:
            if is_classifier is None:
                raise ValueError("'is_classifier' required when task not passed")
        else:
            is_classifier = self.task.classification()

        if is_classifier:
            self.outputs = {
                "probs",
                "logits",
                "pred",
                # 'embeddings',
                # TODO(karan): other information from the model e.g. embeddings which
                #  aren't task related?
            }
        else:
            self.outputs = {
                "pred",
                # 'embeddings',
                # TODO(karan): other information from the model e.g. embeddings which
                #  aren't task related?
            }

        if not device:
            self.device = "cpu"
            if torch.cuda.is_available():
                self.device = "cuda:0"

    def to(self, device: str):
        self.device = device
        return self.model.to(device)

    def __call__(
        self,
        dataset: Dataset,
        input_columns: List[str],
        output_columns: List[str],
        batch_size: int = 32,
        coerce_fn: Callable = None,
        *args,
        **kwargs
    ):

        return self.evaluate(
            dataset,
            input_columns,
            output_columns,
            batch_size,
            coerce_fn,
            *args,
            **kwargs
        )

    @classmethod
    def huggingface(
        cls,
        identifier: str,
        task: Task = None,
        model: Optional[AutoModel] = None,
        tokenizer: Optional[AutoTokenizer] = None,
        is_classifier=None,
    ):
        """

        Args:
            identifier:
            task:
            model:
            tokenizer:

        Returns:

        Examples:
            >>> Model.huggingface(identifier='', task=TernaryNaturalLanguageInference())
            >>> Model.huggingface(identifier='', \
            model=AutoModelForSequenceClassification.from_pretrained(''),
            tokenizer=AutoTokenizer.from_pretrained(''))

        """

        return HuggingfaceModel(
            identifier=identifier,
            task=task,
            model=model,
            tokenizer=tokenizer,
            is_classifier=is_classifier,
        )

    def forward(self, input_batch: Dict) -> Dict:
        raise NotImplementedError

    def evaluate(
        self,
        dataset: Dataset,
        input_columns: List[str],
        output_columns: List[str],
        batch_size: int = 32,
        coerce_fn: Callable = None,
    ):
        raise NotImplementedError

    @staticmethod
    def remap_labels(output_dict: Dict, label_map: List[int]) -> Dict:
        """Map the output labels of the model.

        Example: 3-way classificaiton, with label_map = [1, 2, 0]
        => (model label 0 -> dataset label 1, model label 1 -> dataset label 2, ...).
        """

        # Check the number of classes
        num_classes = len(label_map)

        # Remap the columns of all outputs that have # columns = num_classes
        for key in output_dict:
            if output_dict[key].shape[-1] == num_classes:
                output_dict[key] = output_dict[key][..., label_map]

        # Remap the pred key
        inverse_label_map = [
            t[1] for t in sorted([(label, i) for i, label in enumerate(label_map)])
        ]
        output_dict["pred"] = torch.tensor(inverse_label_map)[output_dict["pred"]]

        return output_dict


class HuggingfaceModel(Model):
    def __init__(
        self,
        identifier: str,
        task: Task = None,
        model: Optional[AutoModel] = None,
        tokenizer: Optional[AutoTokenizer] = None,
        device: str = None,
        is_classifier=None,
    ):

        super(HuggingfaceModel, self).__init__(
            identifier=identifier, task=task, device=device, is_classifier=is_classifier
        )

        self.tokenizer = tokenizer
        if tokenizer is None:
            # Load the tokenizer
            self.tokenizer = AutoTokenizer.from_pretrained(self.identifier)

        self.model = model
        if model is None:
            # Load the model
            if self.task is None:
                if is_classifier is None:
                    raise ValueError("'is_classifier' required when task not specified")
            else:
                is_classifier = self.task.classification()
            if is_classifier:
                self.model = AutoModelForSequenceClassification.from_pretrained(
                    self.identifier
                )
            else:
                self.model = AutoModelForSeq2SeqLM.from_pretrained(self.identifier)

        self.task = task

        # Move the model to device
        self.to(self.device)

    def forward(self, input_batch: Dict) -> Dict:
        # Create the required outputs
        output_dict = {k: None for k in self.outputs}

        if self.task.classification():
            # Run the model on the input_batch
            # TODO(karan): allow outputs to generically contain side information (
            #  embeddings, attention, etc.)
            with torch.no_grad():
                outputs = self.model(**input_batch)

            # The logits are at the 0th index
            logits = outputs[0]

            # TODO(karan): these are still on GPU, do metric computation on GPU then
            #  move to CPU
            # TODO(karan): incrementally compute metrics?
            if "logits" in self.outputs:
                output_dict["logits"] = logits.to("cpu")

            if "probs" in self.outputs:
                output_dict["probs"] = torch.nn.functional.softmax(logits, dim=-1).to(
                    "cpu"
                )

            if "pred" in self.outputs:
                output_dict["pred"] = logits.argmax(dim=-1).to("cpu")
        else:
            with torch.no_grad():
                summary_token_ids = self.model.generate(**input_batch)
                summaries = [
                    self.tokenizer.decode(
                        token_id_list,
                        skip_special_tokens=True,
                        clean_up_tokenization_spaces=False,
                    )
                    for token_id_list in summary_token_ids
                ]
                output_dict["pred"] = summaries

        return output_dict

    def encode_batch(self, batch: Dict[str, List], columns: Collection[str], **kwargs):
        # TODO(karan): Automatically writing this encoder for a variety of tasks
        return self.tokenizer(
            *[batch[key] for key in columns], truncation=True, padding=True, **kwargs
        )

    def predict_batch(self, batch: Dict[str, List], input_columns: Collection[str]):

        # Tokenize the batch
        input_batch = self.encode_batch(batch=batch, columns=input_columns)

        # Convert the batch to torch.Tensor
        input_batch = tz.valmap(
            lambda v: torch.tensor(v).to(device=self.device), input_batch
        )

        # Apply the model to the batch
        return self.forward(input_batch)

    def evaluate(
        self,
        dataset: Dataset,
        input_columns: List[str],
        output_columns: List[str],
        batch_size: int = 32,
        metrics: List[str] = None,
        coerce_fn: Callable = None,
    ):

        # TODO(karan): generalize to TF2

        # Reset the dataset format
        dataset.reset_format()
        dataset.set_format(columns=input_columns + output_columns)

        # TODO(karan): check that the Dataset conforms to the task definition
        # TODO(karan): figure out how the output_columns will be used by the metrics
        pass

        predictions = []
        targets = []

        # Loop and apply the prediction function
        # TODO(karan): not using .map() here in order to get more fine-grained
        #  control over devices
        for idx in range(0, len(dataset), batch_size):
            # Create the batch
            batch = dataset[idx : idx + batch_size]

            # Predict on the batch
            prediction_dict = self.predict_batch(
                batch=batch, input_columns=input_columns
            )

            # Coerce the predictions
            if coerce_fn:
                prediction_dict = coerce_fn(prediction_dict)

            # Grab the raw target key/values
            target_dict = tz.keyfilter(lambda k: k in output_columns, batch)

            # TODO(karan): general version for non-classification problems
            # TODO(karan): move this to the right device
            if self.task.classification():
                target_dict = tz.valmap(lambda v: torch.tensor(v), target_dict)

            # TODO(karan): incremental metric computation here
            # Append the predictions and targets
            predictions.append(prediction_dict)
            targets.append(target_dict)

        # Consolidate the predictions and targets
        if self.task.classification():
            # TODO(karan): Need to store predictions and outputs from the model
            predictions = tz.merge_with(lambda v: torch.cat(v).to("cpu"), *predictions)
            targets = tz.merge_with(lambda v: torch.cat(v).to("cpu"), *targets)
        else:
            predictions = tz.merge_with(
                lambda x: list(itertools.chain.from_iterable(x)), *predictions
            )
            targets = tz.merge_with(
                lambda x: list(itertools.chain.from_iterable(x)), *targets
            )

        # Compute the metrics
        # TODO(karan): generalize this code to support metric computation for any task

        # Assumes classification, so the output_columns contains a single key for the
        # label
        if self.task.classification():
            assert len(output_columns) == 1  # , "Only supports classification."
            num_classes = self.task.output_schema.features[
                list(self.task.output_schema.columns)[0]
            ].num_classes

        labels = targets[list(targets.keys())[0]]

        if metrics is None:
            if self.task is None:
                raise ValueError(
                    "Must specify metrics if model not associated with task"
                )
            metrics = self.task.metrics

        pred = predictions["pred"].to(self.device)
        target = labels.to(self.device)

        evaluation_dict = {
            metric: compute_metric(metric, pred, target, num_classes)
            for metric in metrics
        }

        # Reset the data format
        dataset.reset_format()

        return evaluation_dict

class LudwigModel(Model):
    def __init__(
        self,
    ):
        if not _ludwig_available:
            raise ImportError("Please `pip install ludwig`.")

    def load(
        self,
        model_dir: str,
    ):
        self.ludwig_model = ludwigmodel.load(model_dir)

    def evaluate(
        self,
        dataset: Dataset,
        batch_size: int=128,
        collect_overall_stats: bool=True,
        collect_predictions: bool=True,

    ):
        eval_stats, predictions, _ = self.ludwig_model.evaluate(
                dataset=dataset[:],
                batch_size=batch_size,
                collect_overall_stats=collect_overall_stats,
                collect_predictions=collect_predictions
        )

        return (eval_stats, predictions)


def format_summary(x: str) -> str:
    """Format summary text for computing rouge."""
    re.sub("<n>", "", x)  # remove pegasus newline char
    return "\n".join(nltk.sent_tokenize(x))



from collections import OrderedDict
from typing import List, Union

from datasets.features import ClassLabel, Sequence, Value

from robustnessgym.core.dataset import Dataset
from robustnessgym.core.identifier import Identifier
from robustnessgym.tasks.schema import Schema


class Task:
    """Class for tasks in Robustness Gym."""

    dataset_to_task = {}

    def __init__(
        self,
        identifier: Union[str, Identifier] = "GenericTask",
        input_schema: Schema = None,
        output_schema: Schema = None,
        metrics: List[str] = None,
        *args,
        **kwargs,
    ):
        self._identifier = identifier
        self._input_schema = input_schema
        self._output_schema = output_schema
        self._metrics = metrics

    def __repr__(self):
        return (
            f"task: {self._identifier}\n\n"
            f"Input: {str(self._input_schema)}\n\n"
            f"Output: {str(self._output_schema)}"
        )

    @classmethod
    def lookup(cls, dataset: str):
        return cls.dataset_to_task[dataset]

    @classmethod
    def list_datasets(cls):
        return []

    @property
    def identifier(self):
        """Task identifier."""
        return self._identifier

    @property
    def metrics(self):
        """Task metrics."""
        return self._metrics

    @property
    def input_schema(self):
        """Input schema for the task."""
        return self._input_schema

    @property
    def output_schema(self):
        """Output schema for the task."""
        return self._output_schema

    # @classmethod
    # def from_identifier(cls, identifier):
    #     return getattr()

    @classmethod
    def create(cls, task: str):
        # TODO(karan): figure out how to getattr this
        if task == "TernaryNaturalLanguageInference":
            return TernaryNaturalLanguageInference()
        else:
            raise NotImplementedError

    def remap_schema(self, dataset: Dataset):
        # Ground the schema to the dataset
        input_grounding, reversed_input_grounding = self.input_schema.ground(
            dataset.features
        )
        output_grounding, reversed_output_grounding = self.output_schema.ground(
            dataset.features
        )

        for col in self.input_schema.columns:
            # Grab the column
            values = dataset[input_grounding[col]]
            # Remove it from the dataset
            dataset.remove_column(input_grounding[col])
            # Add again with the right column name
            dataset.add_column(col, values)

        for col in self.output_schema.columns:
            # Grab the column
            values = dataset[output_grounding[col]]
            # Remove it from the dataset
            dataset.remove_column(output_grounding[col])
            # Add again with the right column name
            dataset.add_column(col, values)

        return dataset

        # # Construct a map_fn that remaps the dataset schema
        # def _map_fn(example):
        #     return tz.merge(
        #         {k: example[input_grounding[k]] for k in self.input_schema.features},
        #         {k: example[output_grounding[k]] for k in self.output_schema.features
        #         },
        #     )
        #
        # return dataset.map(
        #     _map_fn,
        #     remove_columns=list(reversed_input_grounding.keys())
        #     + list(reversed_output_grounding.keys()),
        # )

    def classification(self):
        # TODO(karan): improve the schema inference
        # Check that the only output is a ClassLabel output
        if len(self._output_schema) == 1 and isinstance(
            self._output_schema.features[self._output_schema.columns[0]], ClassLabel
        ):
            return True
        return False


class Generic(Task):
    def __init__(self):
        super(Generic, self).__init__(
            identifier="Generic",
            input_schema=None,
            output_schema=None,
            metrics=None,
        )


class SentimentClassification(Task):
    def __init__(self, identifier, input_schema, output_schema, *args, **kwargs):
        super(SentimentClassification, self).__init__(
            identifier=identifier,
            input_schema=input_schema,
            output_schema=output_schema,
            metrics=[
                "accuracy",
                "f1",
                "class_dist",
                "pred_dist"
                # TODO(karan): calibration, other metrics
            ],
            *args,
            **kwargs,
        )


class BinarySentiment(SentimentClassification):
    def __init__(self):
        super(BinarySentiment, self).__init__(
            num_classes=2,
            input_schema=Schema(
                features=OrderedDict(
                    [
                        ("text", Value(dtype="string")),
                    ]
                ),
                grounding_candidates={
                    "text": {"text", "sentence"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict(
                    [
                        ("label", ClassLabel(names=["negative", "positive"])),
                    ]
                ),
                grounding_candidates={
                    "label": {"label"},
                },
            ),
            identifier=self.__class__.__name__,
        )

    @classmethod
    def list_datasets(cls):
        return [
            "imdb",
        ]


class Summarization(Task):
    def __init__(self):
        super(Summarization, self).__init__(
            identifier=self.__class__.__name__,
            input_schema=Schema(
                features=OrderedDict([("text", Value(dtype="string"))]),
                grounding_candidates={
                    "text": {"article", "document"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict([("summary", Value(dtype="string"))]),
                grounding_candidates={
                    "summary": {"highlights", "summary"},
                },
            ),
            metrics=[
                # blah,
                # TODO(karan): calibration, other metrics
                "rouge1",
                "rouge2",
                "rougeLsum",
            ],
        )

    @classmethod
    def list_datasets(cls):
        return [
            "cnn_dailymail",
        ]


class NaturalLanguageInference(Task):
    def __init__(self, identifier, input_schema, output_schema, *args, **kwargs):
        super(NaturalLanguageInference, self).__init__(
            identifier=identifier,
            input_schema=input_schema,
            output_schema=output_schema,
            metrics=[
                "accuracy",
                "f1_micro",
                "f1_macro",
                "class_dist",
                "pred_dist"
                # TODO(karan): calibration, other metrics
            ],
            *args,
            **kwargs,
        )


class BinaryNaturalLanguageInference(NaturalLanguageInference):
    def __init__(self):
        super(BinaryNaturalLanguageInference, self).__init__(
            num_classes=2,
            input_schema=Schema(
                features=OrderedDict(
                    [
                        ("premise", Value(dtype="string")),
                        ("hypothesis", Value(dtype="string")),
                    ]
                ),
                grounding_candidates={
                    "premise": {"premise", "sentence1"},
                    "hypothesis": {"hypothesis", "sentence2"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict(
                    [
                        ("label", ClassLabel(names=["entailment", "non entailment"])),
                    ]
                ),
                grounding_candidates={
                    "label": {"label"},
                },
            ),
            identifier=self.__class__.__name__,
        )

    @classmethod
    def list_datasets(cls):
        return []


class TernaryNaturalLanguageInference(NaturalLanguageInference):
    def __init__(self):
        super(TernaryNaturalLanguageInference, self).__init__(
            num_classes=3,
            input_schema=Schema(
                features=OrderedDict(
                    [
                        ("premise", Value(dtype="string")),
                        ("hypothesis", Value(dtype="string")),
                    ]
                ),
                grounding_candidates={
                    "premise": {"premise", "sentence1"},
                    "hypothesis": {"hypothesis", "sentence2"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict(
                    [
                        (
                            "label",
                            ClassLabel(
                                names=["entailment", "neutral", "contradiction"]
                            ),
                        ),
                    ]
                ),
                grounding_candidates={
                    "label": {"label"},
                },
            ),
            identifier=self.__class__.__name__,
        )

    def datasets(self):
        return {
            "snli",
        }


class QuestionAnswering(Task):
    def __init__(self, identifier, input_schema, output_schema, *args, **kwargs):
        super(QuestionAnswering, self).__init__(
            identifier=identifier,
            input_schema=input_schema,
            output_schema=output_schema,
            metrics=[
                "em",
                "f1",
                # TODO(karan): calibration, other metrics
            ],
            *args,
            **kwargs,
        )


class ExtractiveQuestionAnswering(Task):
    def __init__(self):
        super(ExtractiveQuestionAnswering, self).__init__(
            input_schema=Schema(
                features=OrderedDict(
                    [
                        ("context", Value(dtype="string")),
                        ("question", Value(dtype="string")),
                    ]
                ),
                grounding_candidates={
                    "context": {"context"},
                    "question": {"question"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict(
                    [
                        (
                            "answers",
                            Sequence(
                                feature={
                                    "text": Value(dtype="string", id=None),
                                    "answer_start": Value(dtype="int32", id=None),
                                },
                                length=-1,
                            ),
                        ),
                    ]
                ),
                grounding_candidates={
                    "answers": {
                        "answers",
                    },
                },
            ),
            metrics=[
                "em",
                "f1",
            ],
            identifier=self.__class__.__name__,
        )


# class ExtractiveQuestionAnswering(Task):
#
#     def __init__(self):
#         super(ExtractiveQuestionAnswering, self).__init__(
#             input_schema=Schema(
#                 features=OrderedDict([
#                     ('context', Value(dtype='string')),
#                     ('question', Value(dtype='string')),
#                 ]),
#                 grounding_candidates={
#                     'context': {'context'},
#                     'question': {'question'},
#                 },
#             ),
#             output_schema=Schema(
#                 features=OrderedDict([
#                     ('answer', Sequence(Value(dtype='string'), length=-1)),
#                     ('start', Sequence(Value(dtype='int64'), length=-1)),
#                     ('end', Sequence(Value(dtype='int64'), length=-1)),
#                 ]),
#                 grounding_candidates={
#                     'answer': {
#                         ('answers', 'text'),
#                     },
#                     'start': {
#                         ('answers', 'answer_start')
#                     },
#                     'end': {
#                         lambda answer, start: [idx + len(answer) for idx in start],
#                     },
#                 }
#             ),
#             metrics=[
#                 'em',
#                 'f1',
#             ],
#             identifier=self.__class__.__name__,
#         )

# Evaluation Hierarchy
# --------------------
# (generic task, model) ### QuestionAnswering/NLI
# (narrow task, model) ### MultiHopQuestionAnswering/BinaryNLI
# (dataset, model) ### Particular Dataset/QNLI



"""Easy data augmentation techniques for text classification. Jason Wei and Kai
Zou.

Taken from https://github.com/jasonwei20/eda_nlp
"""

import random
import re
from random import shuffle

try:
    from nltk.corpus import wordnet
except ImportError:
    _nltk_available = False
    wordnet = None
else:
    _nltk_available = True

random.seed(1)

# stop words list
stop_words = [
    "i",
    "me",
    "my",
    "myself",
    "we",
    "our",
    "ours",
    "ourselves",
    "you",
    "your",
    "yours",
    "yourself",
    "yourselves",
    "he",
    "him",
    "his",
    "himself",
    "she",
    "her",
    "hers",
    "herself",
    "it",
    "its",
    "itself",
    "they",
    "them",
    "their",
    "theirs",
    "themselves",
    "what",
    "which",
    "who",
    "whom",
    "this",
    "that",
    "these",
    "those",
    "am",
    "is",
    "are",
    "was",
    "were",
    "be",
    "been",
    "being",
    "have",
    "has",
    "had",
    "having",
    "do",
    "does",
    "did",
    "doing",
    "a",
    "an",
    "the",
    "and",
    "but",
    "if",
    "or",
    "because",
    "as",
    "until",
    "while",
    "of",
    "at",
    "by",
    "for",
    "with",
    "about",
    "against",
    "between",
    "into",
    "through",
    "during",
    "before",
    "after",
    "above",
    "below",
    "to",
    "from",
    "up",
    "down",
    "in",
    "out",
    "on",
    "off",
    "over",
    "under",
    "again",
    "further",
    "then",
    "once",
    "here",
    "there",
    "when",
    "where",
    "why",
    "how",
    "all",
    "any",
    "both",
    "each",
    "few",
    "more",
    "most",
    "other",
    "some",
    "such",
    "no",
    "nor",
    "not",
    "only",
    "own",
    "same",
    "so",
    "than",
    "too",
    "very",
    "s",
    "t",
    "can",
    "will",
    "just",
    "don",
    "should",
    "now",
    "",
]


# cleaning up text


def get_only_chars(line):
    clean_line = ""

    line = line.replace("’", "")
    line = line.replace("'", "")
    line = line.replace("-", " ")  # replace hyphens with spaces
    line = line.replace("\t", " ")
    line = line.replace("\n", " ")
    line = line.lower()

    for char in line:
        if char in "qwertyuiopasdfghjklzxcvbnm ":
            clean_line += char
        else:
            clean_line += " "

    clean_line = re.sub(" +", " ", clean_line)  # delete extra spaces
    if clean_line[0] == " ":
        clean_line = clean_line[1:]
    return clean_line


########################################################################
# Synonym replacement
# Replace n words in the sentence with synonyms from wordnet
########################################################################


def synonym_replacement(words, n):
    new_words = words.copy()
    random_word_list = list(set([word for word in words if word not in stop_words]))
    random.shuffle(random_word_list)
    num_replaced = 0
    for random_word in random_word_list:
        synonyms = get_synonyms(random_word)
        if len(synonyms) >= 1:
            synonym = random.choice(list(synonyms))
            new_words = [synonym if word == random_word else word for word in new_words]
            # print("replaced", random_word, "with", synonym)
            num_replaced += 1
        if num_replaced >= n:  # only replace up to n words
            break

    # this is stupid but we need it, trust me
    sentence = " ".join(new_words)
    new_words = sentence.split(" ")

    return new_words


def get_synonyms(word):
    synonyms = set()
    for syn in wordnet.synsets(word):
        for l in syn.lemmas():
            synonym = l.name().replace("_", " ").replace("-", " ").lower()
            synonym = "".join(
                [char for char in synonym if char in " qwertyuiopasdfghjklzxcvbnm"]
            )
            synonyms.add(synonym)
    if word in synonyms:
        synonyms.remove(word)
    return list(synonyms)


########################################################################
# Random deletion
# Randomly delete words from the sentence with probability p
########################################################################


def random_deletion(words, p):
    # obviously, if there's only one word, don't delete it
    if len(words) == 1:
        return words

    # randomly delete words with probability p
    new_words = []
    for word in words:
        r = random.uniform(0, 1)
        if r > p:
            new_words.append(word)

    # if you end up deleting all words, just return a random word
    if len(new_words) == 0:
        rand_int = random.randint(0, len(words) - 1)
        return [words[rand_int]]

    return new_words


########################################################################
# Random swap
# Randomly swap two words in the sentence n times
########################################################################


def random_swap(words, n):
    new_words = words.copy()
    for _ in range(n):
        new_words = swap_word(new_words)
    return new_words


def swap_word(new_words):
    random_idx_1 = random.randint(0, len(new_words) - 1)
    random_idx_2 = random_idx_1
    counter = 0
    while random_idx_2 == random_idx_1:
        random_idx_2 = random.randint(0, len(new_words) - 1)
        counter += 1
        if counter > 3:
            return new_words
    new_words[random_idx_1], new_words[random_idx_2] = (
        new_words[random_idx_2],
        new_words[random_idx_1],
    )
    return new_words


########################################################################
# Random insertion
# Randomly insert n words into the sentence
########################################################################


def random_insertion(words, n):
    new_words = words.copy()
    for _ in range(n):
        add_word(new_words)
    return new_words


def add_word(new_words):
    synonyms = []
    counter = 0
    while len(synonyms) < 1:
        random_word = new_words[random.randint(0, len(new_words) - 1)]
        synonyms = get_synonyms(random_word)
        counter += 1
        if counter >= 10:
            return
    random_synonym = synonyms[0]
    random_idx = random.randint(0, len(new_words) - 1)
    new_words.insert(random_idx, random_synonym)


########################################################################
# main data augmentation function
########################################################################


def eda(sentence, alpha_sr=0.1, alpha_ri=0.1, alpha_rs=0.1, p_rd=0.1, num_aug=9):
    sentence = get_only_chars(sentence)
    words = sentence.split(" ")
    words = [word for word in words if word != ""]
    num_words = len(words)

    augmented_sentences = []
    num_new_per_technique = int(num_aug / 4) + 1
    n_sr = max(1, int(alpha_sr * num_words))
    n_ri = max(1, int(alpha_ri * num_words))
    n_rs = max(1, int(alpha_rs * num_words))

    # sr
    for _ in range(num_new_per_technique):
        a_words = synonym_replacement(words, n_sr)
        augmented_sentences.append(" ".join(a_words))

    # ri
    for _ in range(num_new_per_technique):
        a_words = random_insertion(words, n_ri)
        augmented_sentences.append(" ".join(a_words))

    # rs
    for _ in range(num_new_per_technique):
        a_words = random_swap(words, n_rs)
        augmented_sentences.append(" ".join(a_words))

    # rd
    for _ in range(num_new_per_technique):
        a_words = random_deletion(words, p_rd)
        augmented_sentences.append(" ".join(a_words))

    augmented_sentences = [get_only_chars(sentence) for sentence in augmented_sentences]
    shuffle(augmented_sentences)

    # trim so that we have the desired number of augmented sentences
    if num_aug >= 1:
        augmented_sentences = augmented_sentences[:num_aug]
    else:
        keep_prob = num_aug / len(augmented_sentences)
        augmented_sentences = [
            s for s in augmented_sentences if random.uniform(0, 1) < keep_prob
        ]

    # append the original sentence
    augmented_sentences.append(sentence)

    return augmented_sentences



from __future__ import annotations

import inspect
import json
import logging
import pathlib
from typing import Callable, Collection, Dict, List, Optional, Sequence, Union

import dill
import pandas as pd
import torch
from fuzzywuzzy import process
from tqdm import tqdm

from robustnessgym.core.constants import (
    ATTACK,
    AUGMENTATION,
    CURATION,
    GENERIC,
    SUBPOPULATION,
)
from robustnessgym.core.dataset import Dataset
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.metrics import compute_metric, get_metric
from robustnessgym.core.model import Model
from robustnessgym.core.report import (
    ClassDistributionColumn,
    NumericColumn,
    Report,
    ScoreColumn,
)
from robustnessgym.core.slice import Slice
from robustnessgym.core.tools import persistent_hash
from robustnessgym.core.version import SemanticVersionerMixin
from robustnessgym.tasks.schema import Schema
from robustnessgym.tasks.task import Task

logger = logging.getLogger(__name__)


class DevBench(SemanticVersionerMixin):
    def __init__(
        self,
        dataset: Dataset,
    ):
        # Call the superclass
        super(DevBench, self).__init__()

        # An identifier for the DevBench
        self.identifier = Identifier("DevBench", dataset=str(dataset.identifier))

        # Dataset that the devbench operates on
        self._dataset = dataset

        # Create the collection of slices
        self._slices = set()
        self._slice_identifiers = set()
        self._slice_table = {}

        # The devbench has aggregators
        self.aggregators = {}

        # The devbench internally tracks metrics
        self.metrics = {}

        # Add slices if any
        self.add_slices(dataset)

    @property
    def dataset(self):
        """Dataset for the devbench."""
        return self._dataset

    @property
    def slices(self):
        """Slices in the devbench."""
        return self._slices

    @property
    def models(self):
        """Models in the devbench."""
        return list(self.aggregators.keys())

    def __repr__(self):
        return f"{self.identifier}(slices={len(self.slices)})"

    def _digest(self) -> str:
        return json.dumps([str(sl) for sl in self.slices])

    def add_slices(
        self,
        slices: Union[Dataset, Slice, Collection[Union[Dataset, Slice]]],
    ) -> None:
        """Add slices to the development bench.

        Args:
            slices: collection of Slice objects
        """
        if isinstance(slices, Slice) or isinstance(slices, Dataset):
            slices = [slices]

        # Add slices
        for sl in slices:
            if not isinstance(sl, Slice) and isinstance(sl, Dataset):
                # Convert `Dataset` to `Slice`
                sl = Slice(sl)

            if isinstance(sl, Slice):
                if (
                    sl.identifier not in self._slice_identifiers
                    and len(sl) > 0
                    and sl.lineage[0][1] == self.dataset.identifier
                ):
                    # Add slices that aren't present, have non-zero length and
                    # originate from the dataset
                    self._slices.add(sl)
                    self._slice_identifiers.add(sl.identifier)
                    self._slice_table[sl.identifier] = sl

        # Calculate all metrics
        self.calculate()

    def add_aggregators(
        self,
        aggregators: Dict[str, Dict[str, Callable]],
    ) -> None:
        """Add functions for aggregation, with a dictionary that maps a model
        name to aggregation functions."""

        # For each model
        for model, aggs in aggregators.items():
            # Iterate over the aggregation functions
            for agg_name, agg in aggs.items():
                assert isinstance(agg, Callable), "Aggregators must be functions."
                arguments = inspect.getfullargspec(agg).args
                assert len(arguments) == 1, "Aggregators must be single argument."
                assert arguments[0] == "dataset", (
                    "Aggregator argument name must be `dataset`, "
                    f"not `{arguments[0]}`."
                )

                # Store the aggregator
                if model not in self.aggregators:
                    self.aggregators[model] = {}
                self.aggregators[model][agg_name] = agg

        # Calculate all metrics
        self.calculate()

    def calculate(self):
        """Calculate all metrics that haven't been calculated yet."""
        # Iterate over all models
        for model in self.aggregators:

            # Add the model to the metrics dict
            if model not in self.metrics:
                self.metrics[model] = {}

            # Iterate over all slices
            for sl in self.slices:
                # Add the slice to the model's metrics dict
                if str(sl.identifier) not in self.metrics[model]:
                    self.metrics[model][str(sl.identifier)] = {}

                # Iterate over all aggregation functions
                for agg_name, aggregator in self.aggregators[model].items():
                    if agg_name not in self.metrics[model][str(sl.identifier)]:
                        # Aggregate
                        self.metrics[model][str(sl.identifier)][agg_name] = aggregator(
                            sl
                        )

    @classmethod
    def from_dataset(
        cls,
        dataset: Dataset,
    ) -> DevBench:
        """Create a DevBench from a dataset."""
        # Create the devbench
        devbench = DevBench(
            dataset=dataset,
        )

        return devbench

    def _human_readable_identifiers(self):
        # Temporary function to generate human readable names
        groups = {}
        for ident in self._slice_identifiers:
            if "->" in str(ident):
                builder_ident = str(ident).split(" -> ")[-1]
                builder_ident, cols = builder_ident.split(" @ ")
                name = builder_ident.split("(")[0]
                if name not in groups:
                    groups[name] = set()
                groups[name].add((builder_ident, cols))

        group_info = {}
        for key, group in groups.items():
            if len(group) == 1:
                group_info[key] = "name"
            else:
                only_single_column = len(set([t[1] for t in group])) == 1
                if only_single_column:
                    group_info[key] = "builder_ident"
                else:
                    group_info[key] = "full"

        ident_mapping = {}
        for ident in self._slice_identifiers:
            if "->" in str(ident):
                builder_ident = str(ident).split(" -> ")[-1]
                builder_ident, cols = builder_ident.split(" @ ")
                name = builder_ident.split("(")[0]

                if group_info[name] == "name":
                    new_ident = name
                elif group_info[name] == "builder_ident":
                    new_ident = builder_ident
                elif group_info[name] == "full":
                    new_ident = str(ident).split(" -> ")[-1]

                if new_ident.startswith("NlpAugTransformation"):
                    new_ident = new_ident.split("NlpAugTransformation(pipeline=[")[
                        1
                    ].split("])")[0]

            else:
                new_ident = str(ident).split("(")[0]

            ident_mapping[ident] = new_ident

        self.ident_mapping = ident_mapping

    def _common_aggregators(self, models: List[str]):

        common_aggregators = []
        # Iterate over all the models
        for model in models:
            assert model in self.metrics, f"Model {model} does not exist."
            common_aggregators.append(set([agg for agg in self.aggregators[model]]))

        # Find the common aggregators by taking an intersection
        return set.intersection(*[set(e) for e in common_aggregators])

    def create_report(
        self,
        models: List[str] = None,
    ) -> Report:
        """Generate report from cached metrics for a model
        Args:
            models: List of models.
        Returns:
            report
        """

        if len(self.slices) == 0:
            raise ValueError("Cannot create report for empty testbench")

        if models is not None:
            for model in models:
                assert model in self.metrics, f"Model {model} does not exist."
        else:
            # Use all the models that are available
            models = list(self.metrics.keys())

        # Set identifiers to be human readable
        self._human_readable_identifiers()

        # Get a list of aggregators that are common to `models`
        common_aggregators = list(self._common_aggregators(models))

        # Populate columns
        columns = []
        for model in models:
            for aggregator in common_aggregators:
                columns.append(
                    ScoreColumn(
                        f"{model}-{aggregator}",
                        min_val=0,
                        max_val=1,
                        is_0_to_1=True,
                    )
                )
        columns.append(NumericColumn("Size"))

        category_names = {
            GENERIC: "Slice",
            SUBPOPULATION: "SubPop",
            ATTACK: "Attack",
            AUGMENTATION: "Augment",
            CURATION: "Eval",
        }

        # Populate data
        data = []
        for sl in self.slices:
            slice_name = self.ident_mapping[sl.identifier]
            slice_size = len(sl)
            slice_category = category_names.get(sl.category, sl.category.capitalize())

            row = [slice_category, slice_name]

            for model in models:
                model_metrics = self.metrics[model]
                if sl.identifier not in model_metrics:
                    continue
                slice_metrics = model_metrics[sl.identifier]
                for agg in common_aggregators:
                    row.append(slice_metrics[agg])

            row.append(slice_size)
            data.append(row)

        df = pd.DataFrame(data)

        report = Report(
            data=df,
            columns=columns,
            dataset_name=str(self.dataset.identifier),
        )
        report.sort(
            category_order=dict(
                (cat, i)
                for i, cat in enumerate(
                    [SUBPOPULATION, AUGMENTATION, CURATION, ATTACK, GENERIC]
                )
            )
        )
        return report

    def search(self, keyword: str, limit: int = 3):
        """Fuzzy search over the slices in the DevBench."""
        return [
            self._slice_table[t[0]]
            for t in process.extract(keyword, self._slice_identifiers, limit=limit)
        ]

    def save(self, path: str) -> None:
        """Save the current devbench to disk. This will save all slices in the
        devbench to disk, as well as metrics and other metadata associated with
        this devbench.

        Args:
            path: string path to the save directory

        Returns: None
        """

        # Path to the save directory
        savedir = pathlib.Path(path) / f"{self.identifier}"

        # Create a directory inside savedir for the slices
        (savedir / "slices").mkdir(parents=True, exist_ok=True)

        # Save all the slices
        self._dataset.save_to_disk(str(savedir / "dataset"))
        pbar = tqdm(self.slices)
        for sl in pbar:
            pbar.set_description(f"Saving slice {str(sl.identifier)[:100]}...")
            sl.save_to_disk(
                str(savedir / "slices" / str(persistent_hash(str(sl.identifier))))
            )

        # Save metrics
        dill.dump(self.metrics, open(str(savedir / "metrics.dill"), "wb"))

        # Save aggregators
        dill.dump(self.aggregators, open(str(savedir / "aggregators.dill"), "wb"))

        # Save metadata
        dill.dump(
            {
                "identifier": self.identifier,
            },
            open(str(savedir / "metadata.dill"), "wb"),
        )

        # Save version info
        with open(str(savedir / "version.dill"), "wb") as f:
            f.write(self._dumps_version())

    @classmethod
    def available(cls, path: str) -> List[str]:
        """Check the list of available devbenches in a directory.

        Args:
            path: string path to a directory. The devbenches available inside this
            directory will be returned.

        Returns: list of available devbenches
        """

        # Path to the save directory
        savedir = pathlib.Path(path)

        # Loop over the folders
        devbench_identifiers = []
        for maybe_devbench in savedir.glob("*"):
            if maybe_devbench.is_dir() and (maybe_devbench / "metadata.dill").exists():
                devbench_identifiers.append(maybe_devbench.name)

        return devbench_identifiers

    @classmethod
    def load(cls, path: str) -> DevBench:
        """Load a devbench from disk.

        Args:
            path: string path to the devbench directory

        Returns:
        """

        # Path to the save directory
        savedir = pathlib.Path(path)

        # Load all the slices
        slices = []
        for sl_path in tqdm(list((savedir / "slices").glob("*"))):
            try:
                slices.append(Slice.load_from_disk(str(sl_path)))
            except FileNotFoundError:
                continue

        # Load dataset
        dataset = Dataset.load_from_disk(str(savedir / "dataset"))

        # Load metrics
        metrics = dill.load(open(str(savedir / "metrics.dill"), "rb"))

        # Load metrics
        aggregators = dill.load(open(str(savedir / "aggregators.dill"), "rb"))

        # Load metadata
        _ = dill.load(open(str(savedir / "metadata.dill"), "rb"))

        # Create the devbench
        devbench = cls(
            dataset=dataset,
        )

        # Set previously stored metrics
        devbench.metrics = metrics

        # Set previously stored aggregators
        devbench.aggregators = aggregators

        # Set the slices
        devbench.add_slices(slices)

        # Load version info
        with open(str(savedir / "version.dill"), "rb") as f:
            devbench._loads_version(f.read())

        return devbench


# TODO(karan): make the TestBench hashable
class TestBench(SemanticVersionerMixin):
    """Class for test benches in Robustness Gym."""

    def __init__(
        self,
        identifier: Union[str, Identifier],
        task: Task = None,
        slices: Collection[Union[Slice, Dataset]] = None,
        version: str = "0.0.1",
        dataset_id: str = None,
        class_names: Collection[str] = None,
    ):

        # Call the superclass
        super(TestBench, self).__init__(version=version)

        # An identifier for the TestBench
        self.identifier = identifier

        # Set the task
        self.task = task
        if task is None:
            self.task = Task()

        # Create the collection of slices
        self._slices = set()
        self.slice_identifiers = set()
        self._slice_table = {}

        # Add slices if any
        if slices:
            self.add_slices(slices)

        # The testbench has calculators
        self.calculators = set()

        # The testbench internally tracks metrics
        self.metrics = {}

        # The schema tells the testbench which columns to extract from the slices for
        # evaluation
        self.schema_type = "default"

        self.dataset_id = dataset_id

        self.class_names = class_names

    @property
    def slices(self):
        """Slices in the testbench."""
        return self._slices

    def __repr__(self):
        return f"TestBench[{self.identifier}](slices={len(self.slices)})"

    def _digest(self) -> str:
        return json.dumps([str(sl) for sl in self.slices])

    @classmethod
    def from_dataset(
        cls,
        dataset: Dataset,
        input_columns: List[str],
        output_columns: List[str],
        # prediction_columns: List[str],
        # metrics: List[str],
    ) -> TestBench:
        """Create a TestBench from a dataset."""
        # Define the task
        task = Task(
            # Identifier
            Identifier("Task", dataset=str(dataset.identifier)),
            # Input and output schemas
            *Schema.for_dataset(dataset, input_columns, output_columns),
        )

        # Create the testbench
        testbench = TestBench(
            identifier=Identifier("TestBench", dataset=str(dataset.identifier)),
            task=task,
            slices=[dataset],
        )

        # testbench.set_single_dataset_mode()
        # testbench.set_prediction_columns(prediction_columns)

        return testbench

    def set_prediction_columns(self, prediction_columns: List[str]):
        """Set the list of columns that act as prediction columns."""
        self.prediction_columns = prediction_columns

    def set_single_dataset_mode(self):
        """All slices must be derived from the root dataset in single dataset
        mode."""
        self.single_dataset_mode = True

    def add_calculators(self, calculators: List[Union[str, Callable]]):
        """Add a list of calculators."""
        for calc in calculators:
            if isinstance(calc, str):
                self.calculators.add(get_metric(calc))
            elif isinstance(calc, Callable):
                self.calculators.add(calc)
            else:
                continue

    def add_model(
        self,
        model: Model = None,
        prediction_columns: List[str] = None,
        identifier: Union[str, Identifier] = None,
    ):

        if prediction_columns is not None:
            assert model is None, "`model` must be None."
            assert (
                identifier is not None
            ), "A model `identifier` must be included with `prediction_columns`."

            self.metrics[identifier] = {}
            for sl in self.slices:
                self.metrics[identifier][sl.identifier] = {}
                for calc in self.calculators:
                    self.metrics[identifier][sl.identifier][calc.__name__] = calc(
                        *[sl[col] for col in prediction_columns],
                        *[sl[col] for col in self.task.output_schema.columns],
                    )

    @classmethod
    def for_dataset(
        cls,
        dataset: str,
        task: Optional[Union[str, Task]] = None,
        version: str = None,
    ):
        """Create a test bench for a dataset."""

        inferred_task = None
        if task is not None:
            # Infer the task from the dataset
            inferred_task = Task.lookup(dataset=dataset)()
            # Check that the inferred task matches the task argument
            if task is not None and task != inferred_task:
                raise AssertionError(
                    f"Dataset {dataset} is only compatible with {inferred_task}, "
                    f"not {task}."
                )

        return TestBench(
            identifier=f"{dataset}-{task}-{version}",
            task=inferred_task,
            slices=[],
        )

    @classmethod
    def for_task(
        cls,
        task: Union[str, Task],
        version: str = None,
    ):
        """Create a testbench for a task."""
        return TestBench(
            identifier=f"{task}-{version}",
            task=task,
            slices=[],
        )

    def _human_readable_identifiers(self):
        # Temporary function to generate human readable names
        groups = {}
        for ident in self.slice_identifiers:
            if "->" in str(ident):
                builder_ident = str(ident).split(" -> ")[-1]
                builder_ident, cols = builder_ident.split(" @ ")
                name = builder_ident.split("(")[0]
                if name not in groups:
                    groups[name] = set()
                groups[name].add((builder_ident, cols))

        group_info = {}
        for key, group in groups.items():
            if len(group) == 1:
                group_info[key] = "name"
            else:
                only_single_column = len(set([t[1] for t in group])) == 1
                if only_single_column:
                    group_info[key] = "builder_ident"
                else:
                    group_info[key] = "full"

        ident_mapping = {}
        for ident in self.slice_identifiers:
            if "->" in str(ident):
                builder_ident = str(ident).split(" -> ")[-1]
                builder_ident, cols = builder_ident.split(" @ ")
                name = builder_ident.split("(")[0]

                if group_info[name] == "name":
                    new_ident = name
                elif group_info[name] == "builder_ident":
                    new_ident = builder_ident
                elif group_info[name] == "full":
                    new_ident = str(ident).split(" -> ")[-1]

                if new_ident.startswith("NlpAugTransformation"):
                    new_ident = new_ident.split("NlpAugTransformation(pipeline=[")[
                        1
                    ].split("])")[0]

            else:
                new_ident = str(ident).split("(")[0]

            ident_mapping[ident] = new_ident

        self.ident_mapping = ident_mapping

    def add_slices(self, slices: Collection[Union[Dataset, Slice]]) -> None:
        """Add slices to the testbench.

        Args:
            slices: collection of Slice objects
        """
        if isinstance(slices, Slice) or isinstance(slices, Dataset):
            slices = [slices]

        # Add slices
        for sl in slices:
            if not isinstance(sl, Slice) and isinstance(sl, Dataset):
                # The slice is a Dataset
                sl = Slice(sl)

            if isinstance(sl, Slice):
                if sl.identifier not in self.slice_identifiers and len(sl) > 0:
                    # Add slices that aren't already present in the testbench and have
                    # non-zero length
                    self.slices.add(sl)
                    self.slice_identifiers.add(sl.identifier)
                    self._slice_table[sl.identifier] = sl

    def evaluate(
        self,
        model: Model,
        batch_size: int = 32,
        coerce_fn: Callable = None,
        input_columns: List[str] = None,
        output_columns: List[str] = None,
    ) -> Dict:
        """Evaluate a model using the test bench and cache results.

        Args:
            model: model to evaluate
            batch_size: batch size for inference
            coerce_fn: function to coerce the model's outputs. Useful if the model's
            outputs cannot directly be compared to the targets.
            input_columns: columns for input schema. Required if task is None.
            output_columns: columns for output schema. Required if task is None.

        Returns: dict mapping slice identifiers to evaluation metrics.
        """

        if self.task is None:
            if input_columns is None or output_columns is None:
                raise ValueError(
                    "Input and output columns required when no task specified."
                )
        else:
            # Set the schema using the task
            # TODO Is the remapping required when not using a task
            self.set_schema("task")
            input_columns = self.task.input_schema.columns
            output_columns = self.task.output_schema.columns

        # TODO(karan): Uncomment and fix this assert on the type of outputs that
        #  model(..) returns
        # # Grab 2 examples from the first slice, run it through the model and check
        # that the output is a dictionary
        # output = model(dataset=Dataset.from_batch(self.slices[0][:2]),
        #                input_keys=self.task.input_schema.keys(),
        #                output_keys=self.task.output_schema.keys(),
        #                batch_size=2,
        #                coerce_fn=coerce_fn)
        # print(output)
        # assert isinstance(output, Sequence) and isinstance(output[0], Mapping), \
        #     "model(..) must return a list of dictionaries. Each dictionary should
        #     map metric names to values."

        # Store the model_metrics
        if model.identifier not in self.metrics:
            self.metrics[model.identifier] = {}

        # Run the model on all the slices
        # TODO(karan): For slices that are subpopulations, the same example can be in
        #  multiple slices
        #  and will be run through the model multiple times. Create a UnionSlice?
        for sl in tqdm(self.slices):
            if sl.identifier not in self.metrics[model.identifier]:
                # Evaluate on the slice
                # TODO Why not update existing results?
                self.metrics[model.identifier][sl.identifier] = model.evaluate(
                    dataset=sl,
                    input_columns=input_columns,
                    output_columns=output_columns,
                    batch_size=batch_size,
                    coerce_fn=coerce_fn,
                )

        return self.metrics[model.identifier]

    def add_predictions(
        self,
        model: Union[Model, str],
        predictions: Dict[str, Union[Sequence, torch.Tensor]],
        output_columns: List[str] = None,
        num_classes=None,
        metrics: List[str] = None,
    ) -> Dict:
        """Compute and cache metrics for pre-computed model predictions
        Args:
            model: Model or model id
            predictions: Map from slice id to sequence or torch Tensor of predictions
            metric (optional): list of metrics. If None, use the metrics specified in
            the task.
            output_columns (optional): names of output columns. Required if testbench
            does not have associated task.
            num_classes (optional): number of classes. Required if testbench does not
            have associated task.
        Returns: computed metrics
        """

        if self.task is None:
            if output_columns is None:
                raise ValueError(
                    "'output_columns' is required if testbench does not have "
                    "associated task."
                )
            if num_classes is None:
                raise ValueError(
                    "'num_classes' is required if testbench does not have associated "
                    "task."
                )
            if metrics is None:
                raise ValueError(
                    "'metrics' is required if testbench does not have associated task."
                )
        else:
            output_columns = self.task.output_schema.columns
            num_classes = self.task.output_schema.features[
                list(self.task.output_schema.columns)[0]
            ].num_classes
            if self.task.classification():
                assert len(output_columns) == 1  # , "Only supports classification."
            if metrics is None:
                metrics = self.task.metrics

        if len(output_columns) > 1:
            raise NotImplementedError("Only single output column supported")

        if isinstance(model, Model):
            model = model.identifier
        if model not in self.metrics:
            self.metrics[model] = {}
        for sl in tqdm(self.slices):
            if sl.identifier not in self.metrics[model]:
                # Evaluate on the slice
                # TODO Why not update existing results?
                # slice_predictions = predictions[sl.identifier]
                evaluation_dict = {}
                # Temporarily expose prediction columns
                # sl.set_format(columns=output_columns()
                # slice_predictions = predictions[sl.identifier]
                # TODO Optimize
                # labels = list(zip(*[sl[col] for col in output_columns]))
                labels = sl[output_columns[0]]
                for metric in metrics:
                    evaluation_dict[metric] = compute_metric(
                        metric=metric,
                        predictions=predictions[sl.identifier],
                        labels=labels,
                        num_classes=num_classes,
                    )
                # sl.reset_format()
                self.metrics[model][sl.identifier] = evaluation_dict

        return evaluation_dict

    def add_metrics(self, model: Union[Model, str], metrics: Dict[str, float]):
        """Cache pre-computed metrics for model
        Args:
            model: Model or model id.
            metrics: map from metric name to value
        """
        if isinstance(model, Model):
            model = model.identifier
        self.metrics[model] = metrics

    def create_report(
        self,
        model: Union[Model, str],
        metric_ids: List[str] = None,
    ) -> Report:
        """Generate report from cached metrics for a model
        Args:
            model: Model or model id. Metrics must have already been computed for
            this model.
            metric_ids (optional): list of metric ids to include in desired order.
            If None, take metrics from sample slice.
        Returns:
            report
        """

        if len(self.slices) == 0:
            raise ValueError("Cannot create report for empty testbench")

        if isinstance(model, Model):
            model = model.identifier
        if model not in self.metrics:
            raise ValueError(
                f"Metrics for model {model} have not been computed yet."
                f" You must first execute one of "
                "the following methods for this model: 'evaluate', "
                "'add_predictions', 'add_metrics'"
            )

        # TODO(Jesse): Need a category for test set

        model_metrics = self.metrics[model]

        # TODO(Jesse): where to put this? Should only need to be called once
        self._human_readable_identifiers()

        if metric_ids is None:
            sample_slice = list(self.slices)[0].identifier
            metric_ids = list(model_metrics[sample_slice].keys())
            sorted_metric_ids = sorted(
                [
                    metric_id
                    for metric_id in metric_ids
                    if metric_id not in ("class_dist", "pred_dist")
                ]
            )
            if "class_dist" in metric_ids:
                sorted_metric_ids.append("class_dist")
            if "pred_dist" in metric_ids:
                sorted_metric_ids.append("pred_dist")
            metric_ids = sorted_metric_ids

        # Populate columns
        columns = []
        for metric_id in metric_ids:
            if metric_id in ("class_dist", "pred_dist"):
                if self.task is None:
                    class_cds = None
                else:
                    class_names = self.task.output_schema.features[
                        list(self.task.output_schema.columns)[0]
                    ].names
                    class_cds = [name[0].upper() for name in class_names]
                columns.append(ClassDistributionColumn(metric_id, class_cds))
            else:
                columns.append(
                    ScoreColumn(metric_id, min_val=0, max_val=1, is_0_to_1=True)
                )
        columns.append(NumericColumn("Size"))

        category_names = {
            GENERIC: "Slice",
            SUBPOPULATION: "SubPop",
            ATTACK: "Attack",
            AUGMENTATION: "Augment",
            CURATION: "Eval",
        }

        # Populate data
        data = []
        for sl in self.slices:
            slice_name = self.ident_mapping[sl.identifier]
            slice_size = len(sl)
            slice_category = category_names.get(sl.category, sl.category.capitalize())
            row = []
            row.append(slice_category)
            row.append(slice_name)
            if sl.identifier not in model_metrics:
                raise ValueError(
                    f"Metrics for model {model} and slice {sl.identifier}"
                    f"have not yet been computed."
                )
            slice_metrics = model_metrics[sl.identifier]
            for metric_id in metric_ids:
                row.append(slice_metrics[metric_id])
            row.append(slice_size)
            data.append(row)

        # TODO(karan): generalize aggregation
        # slice_metrics = tz.merge_with(np.mean, slice_metrics)
        # Task-dependent model predictions
        # TODO(karan): e.g. average class distribution predicted, figure out how to
        #  put this in
        # Task-dependent sl information
        # TODO(karan): e.g. class distribution

        df = pd.DataFrame(data)

        report = Report(
            data=df, columns=columns, model_name=model, dataset_name=self.dataset_id
        )
        report.sort(
            category_order=dict(
                (cat, i)
                for i, cat in enumerate(
                    [SUBPOPULATION, AUGMENTATION, CURATION, ATTACK, GENERIC]
                )
            )
        )
        return report

    def set_schema(self, schema_type: str):
        assert schema_type in {"default", "task"}

        if self.schema_type == schema_type:
            return

        if schema_type == "task":
            self._slices = {self.task.remap_schema(sl) for sl in self.slices}
            self.schema_type = schema_type
        elif schema_type == "default":
            # TODO(karan): undo the schema standardization
            raise NotImplementedError

    def search(self, keyword: str, limit: int = 3):
        return [
            self._slice_table[t[0]]
            for t in process.extract(keyword, self.slice_identifiers, limit=limit)
        ]

    def save(self, path: str) -> None:
        """Save the current testbench to disk. This will save all slices in the
        testbench to disk, as well as metrics and other metadata associated
        with this testbench.

        Args:
            path: string path to the save directory

        Returns: None
        """

        # Path to the save directory
        savedir = pathlib.Path(path) / f"{self.identifier}"

        # Create a directory inside savedir for the slices
        (savedir / "slices").mkdir(parents=True, exist_ok=True)

        # Save all the slices
        pbar = tqdm(self.slices)
        for sl in pbar:
            pbar.set_description(f"Saving slice {str(sl.identifier)[:100]}...")
            sl.save_to_disk(
                str(savedir / "slices" / str(persistent_hash(str(sl.identifier))))
            )

        # Save metrics
        dill.dump(self.metrics, open(str(savedir / "metrics.dill"), "wb"))

        # Save metadata
        dill.dump(
            {
                "task": self.task,
                "identifier": self.identifier,
                "dataset_id": self.dataset_id,
                "slices": [
                    (sl.identifier, sl.category, sl.lineage, len(sl))
                    for sl in self.slices
                ],
            },
            open(str(savedir / "metadata.dill"), "wb"),
        )

        # Save version info
        with open(str(savedir / "version.dill"), "wb") as f:
            f.write(self._dumps_version())

    @classmethod
    def available(cls, path: str) -> List[str]:
        """Check the list of available testbenches in a directory.

        Args:
            path: string path to a directory. The testbenches available inside this
            directory will be returned.

        Returns: list of available testbenches
        """

        # Path to the save directory
        savedir = pathlib.Path(path)

        # Loop over the folders
        testbench_identifiers = []
        for maybe_testbench in savedir.glob("*"):
            if (
                maybe_testbench.is_dir()
                and (maybe_testbench / "metadata.dill").exists()
            ):
                testbench_identifiers.append(maybe_testbench.name)

        return testbench_identifiers

    @classmethod
    def load_metrics(cls, path: str) -> dict:
        """Load metrics from disk."""
        # Path to the save directory
        savedir = pathlib.Path(path)

        # Load metrics
        return dill.load(open(str(savedir / "metrics.dill"), "rb"))

    @classmethod
    def load_metadata(cls, path: str) -> dict:
        """Load metrics from disk."""
        # Path to the save directory
        savedir = pathlib.Path(path)

        # Load metrics
        return dill.load(open(str(savedir / "metadata.dill"), "rb"))

    @classmethod
    def load(cls, path: str) -> TestBench:
        """Load a testbench from disk.

        Args:
            path: string path to the testbench directory

        Returns:
        """

        # Path to the save directory
        savedir = pathlib.Path(path)

        # Load all the slices
        slices = []
        for sl_path in tqdm(list((savedir / "slices").glob("*"))):
            try:
                slices.append(Slice.load_from_disk(str(sl_path)))
            except FileNotFoundError:
                continue

        # Load metrics
        metrics = dill.load(open(str(savedir / "metrics.dill"), "rb"))

        # Load metadata
        metadata = dill.load(open(str(savedir / "metadata.dill"), "rb"))

        # Create the testbench
        testbench = cls(
            identifier=metadata["identifier"],
            task=metadata["task"],
            slices=slices,
        )

        # Set previously stored metrics
        testbench.metrics = metrics

        # Load version info
        with open(str(savedir / "version.dill"), "rb") as f:
            testbench._loads_version(f.read())

        return testbench

    def make(self, identifier: str):
        # Resolve the location of the TestBench

        # Pull the TestBench
        return self.pull(identifier)

    def pull(self, identifier: str):
        pass

    def publish(self):
        pass



from typing import Dict, List, Tuple

import numpy as np

from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.transformation import Transformation
from robustnessgym.slicebuilders.transformations._eda import eda


class EasyDataAugmentation(Transformation):
    """Text transformation class for Easy Data Augmentation.

    Citation
    --------
    Wei, J., & Zou, K. (2019). EDA: Easy Data Augmentation Techniques for Boosting
    Performance on Text Classification
    Tasks. EMNLP 2019.
    """

    def __init__(
        self, num_transformed=1, alpha_sr=0.1, alpha_ri=0.1, alpha_rs=0.1, p_rd=0.1
    ):

        super(EasyDataAugmentation, self).__init__(
            identifiers=Identifier.range(
                n=num_transformed,
                _name=self.__class__.__name__,
                alpha_sr=alpha_sr,
                alpha_ri=alpha_ri,
                alpha_rs=alpha_rs,
                p_rd=p_rd,
            )
        )

        # Set the parameters
        self.alpha_sr = alpha_sr
        self.alpha_ri = alpha_ri
        self.alpha_rs = alpha_rs
        self.p_rd = p_rd

        # Download wordnet
        self._download_wordnet()

    def _download_wordnet(self):
        try:
            if not (self.logdir / "wordnet").exists():
                import nltk

                nltk.download(
                    "wordnet", download_dir=str(self.logdir / "wordnet"), quiet=False
                )
        except ImportError:
            print("Need nltk.")

    def apply(
        self,
        skeleton_batches: List[Dict[str, List]],
        slice_membership: np.ndarray,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs
    ) -> Tuple[List[Dict[str, List]], np.ndarray]:

        for col in columns:
            # Iterate over col for all examples in the batch
            for i, text in enumerate(batch[col]):
                try:
                    # EDA returns a list of augmented text, including the original
                    # text at the last position
                    augmented_texts = eda(
                        text,
                        alpha_sr=self.alpha_sr,
                        alpha_ri=self.alpha_ri,
                        alpha_rs=self.alpha_rs,
                        p_rd=self.p_rd,
                        num_aug=self.num_transformed,
                    )[:-1]

                    # Store the augmented text in the augmented batches
                    for j, augmented_text in enumerate(augmented_texts):
                        skeleton_batches[j][col][i] = augmented_text
                except:  # noqa
                    # Unable to augment the example: set its slice membership to zero
                    slice_membership[i, :] = 0

        return skeleton_batches, slice_membership



# from .allen.allen_predictor import AllenPredictor
# from .allen.constituency_parser.constituency_parser import AllenConstituencyParser
# from .allen.dependency_parser.dependency_parser import AllenDependencyParser
# from .allen.semantic_role_labeler.semantic_role_labeler import
# AllenSemanticRoleLabeler
# from .bootleg.bootleg import Bootleg
# from .cached_ops import (
#     CachedOperation,
#     stow
# )
# from .custom.strip_text import StripText
# from .spacy.spacy import Spacy
# from .textblob.textblob import TextBlob
from .spacy import Spacy  # noqa
from .stanza import Stanza  # noqa
from .textblob import TextBlob  # noqa



"""Implementation of the Operation abstract base class."""
import json
from abc import ABC, abstractmethod
from typing import Callable, List

from robustnessgym.core.identifier import Identifier
from robustnessgym.core.tools import persistent_hash


class Operation(ABC):
    """Abstract base class for operations in Robustness Gym."""

    def __init__(
        self,
        apply_fn: Callable = None,
        identifiers: List[Identifier] = None,
        num_outputs: int = None,
        *args,
        **kwargs
    ):

        if not identifiers:
            assert (
                num_outputs
            ), "Must pass in num_outputs if no identifiers are specified."

        # Set the identifiers for the outputs of the Operation
        self._identifiers = (
            Identifier.range(n=num_outputs, _name=self.__class__.__name__, **kwargs)
            if not identifiers
            else identifiers
        )

        # Assign the apply_fn
        if apply_fn:
            self.apply = apply_fn

        # # Find the batch and dataset processors
        # self._batch_processors = {method.__name__ for method in
        #                           methods_with_decorator(self.__class__,
        #                           batch_processing)}
        # self._dataset_processors = {method.__name__ for method in
        #                             methods_with_decorator(self.__class__,
        #                             dataset_processing)}

    @property
    def identifiers(self):
        return self._identifiers

    # @property
    # @abstractmethod
    # def processors(self):
    #     raise NotImplementedError("Must specify the order in which processors are
    #     applied.")
    #
    # @property
    # def batch_processors(self):
    #     return self._batch_processors
    #
    # @property
    # def dataset_processors(self):
    #     return self._dataset_processors

    def __hash__(self):
        """Compute a hash value for the cached operation object."""
        val = 0
        for identifier in self.identifiers:
            val ^= persistent_hash(str(identifier))
        return val

    # def get_cache_hash(self,
    #                    columns: List[str],
    #                    processor: str = None):
    #     """
    #     Construct a hash that will be used to identify the application of a
    #     Operation to the columns of a dataset.
    #     """
    #
    #     # Hash the Operation
    #     val = hash(self)
    #
    #     # Combine with the hash for each column
    #     for column in columns:
    #         val ^= persistent_hash(column)
    #
    #     # Combine with the hash for the processor
    #     if processor:
    #         val ^= persistent_hash(processor)
    #
    #     return val
    #
    # def get_cache_file_name(self,
    #                         columns: List[str],
    #                         processor: str = None) -> str:
    #     """
    #     Construct a file name for caching.
    #     """
    #     return 'cache-' + str(abs(self.get_cache_hash(columns=columns,
    #     processor=processor))) + '.arrow'

    # # FIXME: temporary
    # def __call__(self,
    #              batch_or_dataset: BatchOrDataset,
    #              columns: List[str],
    #              mask: List[int] = None,
    #              *args,
    #              **kwargs) -> BatchOrDataset:
    #
    #     if isinstance(batch_or_dataset, Dataset):
    #         # Check the Dataset's InteractionTape to see if the Operation was
    #         previously applied
    #         if not mask:
    #             # This infers a mask that specifies which outputs of the Operation
    #             are not required
    #             mask = batch_or_dataset.check_tape(
    #                 path=[self.__class__.__name__],
    #                 identifiers=self.identifiers,
    #                 columns=columns
    #             )
    #
    #         # If all outputs of the Operation were previously present in the
    #         Dataset, simply return
    #         if all(mask):
    #             return batch_or_dataset
    #
    #         # Apply the CachedOperation to the dataset
    #         dataset = self.process_dataset(
    #             dataset=batch_or_dataset,
    #             columns=columns,
    #         )
    #
    #         # Update the InteractionTape with the applied CachedOperation
    #         dataset.update_tape(
    #             path=[CACHED_OPS],
    #             identifiers=self.identifiers,
    #             columns=columns,
    #         )
    #
    #         return dataset
    #
    #     elif isinstance(batch_or_dataset, Dict):
    #
    #         assert len(self.dataset_processors) == 0, \
    #             f"Cannot apply {self.__class__.__name__} to a batch, " \
    #             f"since it has dataset processors: {self.dataset_processors}. " \
    #             f"Use Dataset.from_batch(batch) before calling {
    #             self.__class__.__name__}."
    #
    #         # Apply the Operation
    #         return self.process_batch(
    #             batch=batch_or_dataset,
    #             columns=columns
    #         )
    #     else:
    #         raise NotImplementedError
    #
    # def wrap_batch_processor(self,
    #                          batch_processor: Callable) -> Callable:
    #
    #     def _wrap_batch_processor(batch: Batch,
    #                               columns: List[str],
    #                               **kwargs):
    #
    #         return batch_processor(batch=batch, columns=columns, **kwargs)
    #
    #     return _wrap_batch_processor
    #
    # def process_dataset(self,
    #                     dataset: Dataset,
    #                     columns: List[str],
    #                     batch_size: int = 32) -> Dataset:
    #     """
    #     Apply the Operation to a dataset.
    #     """
    #
    #     # Apply them in order
    #     for method in self.processors:
    #
    #         # Apply batch processors by .map(..) over the dataset
    #         if method.__name__ in self.batch_processors:
    #             dataset = dataset.map(
    #                 partial(method, columns=columns),
    #                 batched=True,
    #                 batch_size=batch_size,
    #                 cache_file_name=self.get_cache_file_name(columns=columns,
    #                 processor=method)
    #             )
    #         # Apply dataset processors directly
    #         elif method.__name__ in self.dataset_processors:
    #             dataset = method(
    #                 dataset=dataset,
    #                 columns=columns,
    #             )
    #         else:
    #             raise RuntimeError(f"{method} is not a processor. "
    #                                f"Please remove {method} from the `processors`
    #                                property or decorate it.")
    #
    #     return dataset
    #
    # def process_batch(self,
    #                   batch: Batch,
    #                   columns: List[str]) -> Batch:
    #     """
    #     Apply the cached operation to a batch.
    #     """
    #     assert len(set(columns) - set(batch.keys())) == 0, "Any column in 'columns'
    #     must be present in 'batch'."
    #
    #     # Run the cached operation, and encode outputs (defaults to json.dumps)
    #     encoded_outputs = [
    #         self.encode(example_output)
    #         for example_output in self.apply(batch=batch, columns=columns)
    #     ]
    #
    #     # Construct updates
    #     updates = self.construct_updates(
    #         encoded_outputs=encoded_outputs,
    #         columns=columns
    #     )
    #
    #     # Update the cache and return the updated batch
    #     return self.store(batch=batch, updates=updates)

    @classmethod
    def identify(cls, **kwargs):
        return Identifier(_name=cls.__name__, **kwargs)

    @classmethod
    def encode(cls, obj) -> str:
        """

        Args:
            obj:

        Returns:

        """
        return json.dumps(obj)

    @classmethod
    def decode(cls, s: str):
        """

        Args:
            s:

        Returns:

        """
        return json.loads(s)

    @abstractmethod
    def apply(self, *args, **kwargs):
        """

        Args:
            *args:
            **kwargs:

        Returns:

        """
        pass



from robustnessgym.cachedops.allen.allen_predictor import AllenPredictor


class AllenSemanticRoleLabeler(AllenPredictor):
    def __init__(self, device: str = None):
        super(AllenSemanticRoleLabeler, self).__init__(
            path="https://storage.googleapis.com/allennlp-public-models/bert-base-srl"
            "-2020.03.24.tar.gz",
            device=device,
        )



from robustnessgym.cachedops.allen.allen_predictor import AllenPredictor


class AllenDependencyParser(AllenPredictor):
    def __init__(self, device: str = None):
        super(AllenDependencyParser, self).__init__(
            path="https://storage.googleapis.com/allennlp-public-models/"
            "biaffine-dependency-parser-ptb-2020.04.06.tar.gz",
            device=device,
        )



"""Generic wrapper for adversarial attacks."""
from typing import List

from robustnessgym.core.constants import ATTACK
from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.transformation import Transformation


class Attack(Transformation):
    """Class for adversarial attacks."""

    def __init__(
        self,
        identifiers: List[Identifier],
        apply_fn=None,
    ):
        super(Attack, self).__init__(
            category=ATTACK,
            identifiers=identifiers,
            apply_fn=apply_fn,
        )



from __future__ import annotations

from typing import Any, Dict, List, Sequence

import cytoolz as tz
import numpy as np
from ahocorasick import Automaton

from robustnessgym.cachedops.spacy import Spacy
from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.subpopulation import Subpopulation


class AhoCorasick:
    def __init__(self, *args, **kwargs):
        super(AhoCorasick, self).__init__(*args, **kwargs)

        # Use the Aho-Corasick search algorithm to speed up phrase lookups
        self.automaton = Automaton()

    @classmethod
    def from_phrases(cls, phrases: Dict[Any, str]) -> AhoCorasick:
        # Create a new automaton
        ahocorasick = cls()

        # Add all the phrases we want to search for
        for key, phrase in phrases.items():
            # As values, we add the key of the phrase
            ahocorasick.automaton.add_word(phrase, key)

        # Initialize Aho-Corasick
        ahocorasick.automaton.make_automaton()

        return ahocorasick


class HasPhrase(
    Subpopulation,
    # Spacy
):
    def __init__(
        self, phrases=None, identifiers: List[Identifier] = None, *args, **kwargs
    ):

        super(HasPhrase, self).__init__(
            # One slice per phrase
            identifiers=[
                Identifier(_name=self.__class__.__name__, phrase=phrase)
                for phrase in phrases
            ]
            if not identifiers
            else identifiers,
            *args,
            **kwargs
        )

        # This is the list of phrases that will be searched
        self.phrases = phrases
        if self.phrases is None:
            self.phrases = []

        # Create and populate Aho-Corasick automatons for words and phrases
        self.word_ahocorasick = AhoCorasick.from_phrases(
            {i: phrase for i, phrase in enumerate(self.phrases) if " " not in phrase}
        )
        self.phrase_ahocorasick = AhoCorasick.from_phrases(
            {i: phrase for i, phrase in enumerate(self.phrases) if " " in phrase}
        )

    @classmethod
    def from_file(cls, path: str) -> Subpopulation:
        """Load phrases from a file, one per line."""
        with open(path) as f:
            phrases = [line.strip() for line in f.readlines()]
        return cls(phrases=phrases)

    @classmethod
    def default(cls) -> Subpopulation:
        """A default vocabulary of phrases to search."""
        return cls(phrases=[])

    @classmethod
    def join(cls, *slicers: HasPhrase) -> Sequence[HasPhrase]:
        """Join to combine multiple HasPhrase slicers."""
        return [
            HasPhrase(phrases=list(tz.concat([slicer.phrases for slicer in slicers])))
        ]

    def apply(
        self,
        slice_membership: np.ndarray,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs
    ) -> np.ndarray:

        # Use the spacy cache to grab the tokens in each example (for each key)
        tokenized_batch = Spacy.retrieve(
            batch=batch,
            columns=[[key] for key in columns],
            proc_fns="tokens",
        )

        # Search for words
        if len(self.word_ahocorasick.automaton) > 0:
            for key, tokens_batch in tokenized_batch.items():
                for i, tokens in enumerate(tokens_batch):
                    # Get the values (indices) of all the matched tokens
                    matched_indices = [
                        self.word_ahocorasick.automaton.get(token)
                        for token in tokens
                        if self.word_ahocorasick.automaton.exists(token)
                    ]

                    # Fill in the slice labels for slices that are present
                    slice_membership[i, matched_indices] = 1

        # Search for phrases
        if len(self.phrase_ahocorasick.automaton) > 0:
            for key in columns:
                for i, example in enumerate(batch[key]):
                    # Get the values (indices) of all the matched phrases
                    matched_indices = [
                        index
                        for _, index in self.phrase_ahocorasick.automaton.iter(example)
                    ]

                    # Fill in the slice labels for slices that are present
                    slice_membership[i, matched_indices] = 1

        return slice_membership


# class HasAnyPhrase(Subpopulation):
#
#     def __init__(self,
#                  phrases: List[str] = None,
#                  identifier: Identifier = None,
#                  *args,
#                  **kwargs):
#         # Take the union of the phrases
#         subpopulation = Subpopulation.union(
#             HasPhrase(phrases=phrases),
#             identifier=Identifier(
#                 _name=self.__class__.__name__,
#                 phrases=set(phrases),
#             ) if not identifier else identifier
#         )
#
#         super(HasAnyPhrase, self).__init__(
#             identifiers=subpopulation.identifiers,
#             apply_fn=subpopulation.apply,
#             *args,
#             **kwargs
#         )


class HasAnyPhrase(Subpopulation):
    def __init__(
        self,
        phrase_groups: List[List[str]] = None,
        identifiers: List[Identifier] = None,
        *args,
        **kwargs
    ):

        # Keep track of the phrase groups
        self.phrase_groups = phrase_groups

        if identifiers:
            assert len(identifiers) == len(
                phrase_groups
            ), "Must have one identifier per phrase group."

        self.subpopulations = []
        # For every phrase group
        for i, phrases in enumerate(phrase_groups):
            # Take the union of the phrases
            self.subpopulations.append(
                Subpopulation.union(
                    HasPhrase(phrases=phrases),
                    identifier=Identifier(
                        _name=self.__class__.__name__,
                        phrases=set(phrases),
                    )
                    if not identifiers
                    else identifiers[i],
                )
            )

        super(HasAnyPhrase, self).__init__(
            identifiers=list(
                tz.concat(
                    [subpopulation.identifiers for subpopulation in self.subpopulations]
                )
            ),
            *args,
            **kwargs
        )

    def apply(
        self,
        slice_membership: np.ndarray,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs
    ) -> np.ndarray:

        # Run all the subpopulations in sequence to update the slice membership matrix
        for i, subpopulation in enumerate(self.subpopulations):
            slice_membership[:, i : i + 1] = subpopulation.apply(
                slice_membership=slice_membership[:, i : i + 1],
                batch=batch,
                columns=columns,
                *args,
                **kwargs
            )

        return slice_membership

    @classmethod
    def join(cls, *slicebuilders: HasAnyPhrase) -> Sequence[HasAnyPhrase]:
        # Join all the slicebuilders
        return [
            HasAnyPhrase(
                phrase_groups=[
                    phrases
                    for slicebuilder in slicebuilders
                    for phrases in slicebuilder.phrase_groups
                ],
                identifiers=[
                    identifier
                    for slicebuilder in slicebuilders
                    for identifier in slicebuilder.identifiers
                ],
            )
        ]


class HasAllPhrases(Subpopulation):
    def __init__(
        self,
        phrase_groups: List[List[str]] = None,
        identifiers: List[Identifier] = None,
        *args,
        **kwargs
    ):

        # Keep track of the phrase groups
        self.phrase_groups = phrase_groups

        if identifiers:
            assert len(identifiers) == len(
                phrase_groups
            ), "Must have one identifier per phrase group."

        self.subpopulations = []
        # For every phrase group
        for i, phrases in enumerate(phrase_groups):
            # Take the union of the phrases
            self.subpopulations.append(
                Subpopulation.intersection(
                    HasPhrase(phrases=phrases),
                    identifier=Identifier(
                        _name=self.__class__.__name__,
                        phrases=set(phrases),
                    )
                    if not identifiers
                    else identifiers[i],
                )
            )

        super(HasAllPhrases, self).__init__(
            identifiers=list(
                tz.concat(
                    [subpopulation.identifiers for subpopulation in self.subpopulations]
                )
            ),
            *args,
            **kwargs
        )

    def apply(
        self,
        slice_membership: np.ndarray,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs
    ) -> np.ndarray:

        # Run all the subpopulations in sequence to update the slice membership matrix
        for i, subpopulation in enumerate(self.subpopulations):
            slice_membership[:, i : i + 1] = subpopulation.apply(
                slice_membership=slice_membership[:, i : i + 1],
                batch=batch,
                columns=columns,
                *args,
                **kwargs
            )

        return slice_membership

    @classmethod
    def join(cls, *slicebuilders: HasAllPhrases) -> Sequence[HasAllPhrases]:
        # Join all the slicebuilders
        return [
            HasAllPhrases(
                phrase_groups=[
                    phrases
                    for slicebuilder in slicebuilders
                    for phrases in slicebuilder.phrase_groups
                ],
                identifiers=[
                    identifier
                    for slicebuilder in slicebuilders
                    for identifier in slicebuilder.identifiers
                ],
            )
        ]


# class HasAllPhrases(Subpopulation):
#
#     def __init__(self,
#                  phrases=None,
#                  identifier: Identifier = None,
#                  *args,
#                  **kwargs):
#         # Take the intersection of the phrases
#         subpopulation = Subpopulation.intersection(
#             HasPhrase(phrases=phrases),
#             identifier=Identifier(
#                 _name=self.__class__.__name__,
#                 phrases=set(phrases),
#             ) if not identifier else identifier,
#         )
#
#         super(HasAllPhrases, self).__init__(
#             identifiers=subpopulation.identifiers,
#             apply_fn=subpopulation.apply,
#             *args,
#             **kwargs
#         )


class HasIndefiniteArticle(HasAnyPhrase):
    def __init__(self):
        super(HasIndefiniteArticle, self).__init__(
            phrase_groups=[["a", "an"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HasDefiniteArticle(HasAnyPhrase):
    def __init__(self):
        super(HasDefiniteArticle, self).__init__(
            phrase_groups=[["the"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HasTemporalPreposition(HasAnyPhrase):
    def __init__(self):
        super(HasTemporalPreposition, self).__init__(
            phrase_groups=[["after", "before", "past"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HasPosessivePreposition(HasAnyPhrase):
    def __init__(self):
        super(HasPosessivePreposition, self).__init__(
            phrase_groups=[["inside of", "with", "within"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HasComparison(HasAnyPhrase):
    def __init__(self):
        super(HasComparison, self).__init__(
            phrase_groups=[["more", "less", "better", "worse", "bigger", "smaller"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HasQuantifier(HasAnyPhrase):
    def __init__(self):
        super(HasQuantifier, self).__init__(
            phrase_groups=[["all", "some", "none"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HasNegation(HasAnyPhrase):
    def __init__(self):
        super(HasNegation, self).__init__(
            phrase_groups=[
                [
                    "no",
                    "not",
                    "none",
                    "noone ",
                    "nobody",
                    "nothing",
                    "neither",
                    "nowhere",
                    "never",
                    "hardly",
                    "scarcely",
                    "barely",
                    "doesnt",
                    "isnt",
                    "wasnt",
                    "shouldnt",
                    "wouldnt",
                    "couldnt",
                    "wont",
                    "cant",
                    "dont",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )



from __future__ import annotations

import json
from copy import copy
from json import JSONDecodeError

from robustnessgym.core.constants import CURATION
from robustnessgym.core.dataset import Dataset
from robustnessgym.core.identifier import Identifier


class Slice(Dataset):
    """Slice class in Robustness Gym."""

    def __init__(
        self,
        *args,
        identifier: Identifier = None,
        **kwargs,
    ):
        # Set the identifier
        self._identifier = identifier

        # A slice has a lineage
        self.lineage = []

        # Set the category of the slice: defaults to 'curated'
        self.category = CURATION

        # Create a Slice directly from a Dataset
        if len(args) == 1 and isinstance(args[0], Dataset):
            # Update with the dataset info
            dataset = args[0]
            self.__dict__.update(dataset.__dict__)
            self._dataset = copy(dataset._dataset)
            # self._identifier = identifier or dataset.identifier
            self.lineage = [(str(Dataset.__name__), dataset.identifier)]
        else:
            super(Slice, self).__init__(*args, **kwargs)

    def __repr__(self):
        return (
            f"RG{self.__class__.__name__}["
            f"num_rows: {self.num_rows}]({self.identifier})"
        )

    def add_to_lineage(self, category, identifier, columns=None):
        """Append to the lineage."""
        if columns:
            self.lineage.append((category, identifier, columns))
        else:
            self.lineage.append((category, identifier))

        # Update the identifier
        self._lineage_to_identifier()

    def _lineage_to_identifier(self):
        """Synchronize to the current lineage by reassigning to
        `self._identifier`."""
        short_lineage = []
        for entry in self.lineage:
            if len(entry) == 3:
                try:
                    columns = json.loads(entry[2])
                except JSONDecodeError:
                    columns = entry[2]
                short_lineage.append(str(entry[1]) + " @ " + str(columns))
            else:
                short_lineage.append(str(entry[1]))
        # Assign the new lineage to the identifier
        self._identifier = Identifier(_name=" -> ".join(short_lineage))

    @property
    def identifier(self):
        """Slice identifier."""
        if self._identifier:
            return self._identifier
        if self.lineage:
            self._lineage_to_identifier()
            return self._identifier
        return None

    @identifier.setter
    def identifier(self, value):
        """Set the slice's identifier."""
        self._identifier = value

    @classmethod
    def from_dataset(cls, dataset: Dataset):
        """Create a slice from a dataset."""
        return cls(dataset)

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        dataset_state_keys = super(Slice, cls)._state_keys()
        return dataset_state_keys.union(
            {
                "lineage",
                "category",
            }
        )

    def __getstate__(self):
        """Get the current state of the slice."""

        dataset_state = super(Slice, self).__getstate__()
        state = {
            **dataset_state,
            **{
                "lineage": [
                    tuple(t[:1])
                    + (t[1].dumps(),)
                    + (tuple(t[2:]) if len(t) > 2 else ())
                    for t in self.lineage
                ],
                "category": self.category,
            },
        }
        self._assert_state_keys(state)

        return state

    def __setstate__(self, state):
        """Set the current state of the slice."""
        # Check that the state contains all keys
        self._assert_state_keys(state)

        # Load the lineage
        self.lineage = [
            tuple(t[:1])
            + (Identifier.loads(t[1]),)
            + (tuple(t[2:]) if len(t) > 2 else ())
            for t in state["lineage"]
        ]

        # Load the category
        self.category = state["category"]

        # Set the dataset state: pick out state keys that correspond to the Dataset
        super(Slice, self).__setstate__(
            {k: v for k, v in state.items() if k in super(Slice, self)._state_keys()}
        )



from __future__ import annotations

from typing import Dict, List, Tuple

import numpy as np

from robustnessgym.slicebuilders.transformation import Transformation


class GPT3Transformation(Transformation):
    def __init__(
        self,
    ):
        super(GPT3Transformation, self).__init__()

    def apply(
        self,
        skeleton_batches: List[Dict[str, List]],
        slice_membership: np.ndarray,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs
    ) -> Tuple[List[Dict[str, List]], np.ndarray]:
        pass



from typing import List

from robustnessgym.core.cachedops import SingleColumnCachedOperation


class StripText(SingleColumnCachedOperation):
    def __init__(self):
        super(StripText, self).__init__()

    def single_column_apply(self, column_batch: List, *args, **kwargs) -> List:
        # Clean up each text with a simple function and return the stripped text
        return list(
            map(
                lambda text: text.lower()
                .replace(".", "")
                .replace("?", "")
                .replace("!", "")
                .replace(",", ""),
                column_batch,
            )
        )



from typing import Dict, List, Tuple

import numpy as np

from robustnessgym.core.dataset import transpose_batch
from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.attack import Attack

try:
    from morpheus import MorpheusHuggingfaceNLI, MorpheusHuggingfaceQA
except ImportError:
    _morpheus_available = False
else:
    _morpheus_available = True


class Morpheus(Attack):
    def __init__(self, dataset: str, model: str, constrain_pos: bool = True, **kwargs):

        if not _morpheus_available:
            raise ImportError("Please install morpheus.")

        super().__init__(
            identifiers=[
                Identifier(
                    self.__class__.__name__,
                    dataset=dataset,
                    model=model,
                )
            ],
        )

        self.constrain_pos = constrain_pos

        self.dataset = dataset.lower()
        if self.dataset == "mnli":
            self.attack = MorpheusHuggingfaceNLI(model)
        elif "squad" in self.dataset:
            is_squad2 = "2" in self.dataset
            self.attack = MorpheusHuggingfaceQA(model, squad2=is_squad2)
        elif self.dataset == "cnn_dailymail" or self.dataset == "xsum":
            rouge_type = kwargs.get("rouge_type", "rougeL")
            max_input_tokens = kwargs.get("max_input_tokens", 1024)
            self.attack = MorpheusHuggingfaceQA(
                model, rouge_type=rouge_type, max_input_tokens=max_input_tokens
            )
        else:
            raise NotImplementedError

    def apply(
        self,
        skeleton_batches: List[Dict[str, List]],
        slice_membership: np.ndarray,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs
    ) -> Tuple[List[Dict[str, List]], np.ndarray]:

        for i, example in enumerate(transpose_batch(batch)):
            if self.dataset == "mnli":
                # Assume column order is [premise, hypothesis, label]
                prem_col, hypo_col, label_col = columns
                text_label = self.get_NLI_text_label(example[label_col])

                new_prem, new_hypo, predicted_label, _ = self.attack.morph(
                    example[prem_col],
                    example[hypo_col],
                    example[label_col],
                    constrain_pos=self.constrain_pos,
                )
                if predicted_label != text_label:
                    skeleton_batches[0][prem_col][i] = new_prem
                    skeleton_batches[0][hypo_col][i] = new_hypo
                else:
                    slice_membership[i, 0] = 0
            elif "squad" in self.dataset:
                question_col = columns[0]
                # NOTE: assume first element in columns is question_col
                # Ignoring the rest since example['answers'] is another Dict
                question_dict = self.prepare_question_dict(example, question_col)
                new_question, predicted_answer = self.attack.morph(
                    question_dict, example["context"], constrain_pos=self.constrain_pos
                )
                if predicted_answer not in example["answers"]["text"]:
                    skeleton_batches[0][question_col][i] = new_question
                else:
                    slice_membership[i, 0] = 0
            elif self.dataset == "cnn_dailymail" or self.dataset == "xsum":
                # Assume column order is [article_col, summary_col]
                article_col, summary_col = columns
                new_article, predicted_summary, _ = self.attack.morph(
                    example[article_col],
                    example[summary_col],
                    constrain_pos=self.constrain_pos,
                )
                if predicted_summary != example[summary_col]:
                    skeleton_batches[0][article_col][i] = new_article
                else:
                    slice_membership[i, 0] = 0
            else:
                raise NotImplementedError
        return skeleton_batches, slice_membership

    # No type hint since the values can be ints or strings: Dict[str,]
    @classmethod
    def prepare_question_dict(cls, example, question_col):
        question_dict = {"question": example[question_col]}
        question_dict["answers"] = [
            {"answer_start": i[0], "text": i[1]}
            for i in zip(example["answers"]["answer_start"], example["answers"]["text"])
        ]
        question_dict["is_impossible"] = len(example["answers"]["text"]) == 0
        return question_dict

    def get_NLI_text_label(self, label: int) -> str:
        hf_labels = ["entailment", "neutral", "contradiction"]
        return hf_labels[label]



from __future__ import annotations

import pathlib

# from functools import partial
from itertools import compress
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import cytoolz as tz
import numpy as np
import tqdm

from robustnessgym.core.constants import (
    ATTACK,
    CURATION,
    GENERIC,
    SLICEBUILDERS,
    SUBPOPULATION,
    TRANSFORMATION,
)
from robustnessgym.core.dataset import Batch, BatchOrDataset, Dataset
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.slice import Slice
from robustnessgym.core.storage import StorageMixin
from robustnessgym.core.tools import recmerge, strings_as_json

# from multiprocess.pool import Pool


class SliceBuilder(StorageMixin):
    """Base class for builders that output slices."""

    # Path to a log directory
    logdir: pathlib.Path = pathlib.Path.home() / "robustnessgym" / SLICEBUILDERS

    # Create the log directory
    logdir.mkdir(parents=True, exist_ok=True)

    CATEGORIES = [
        GENERIC,
        SUBPOPULATION,
        ATTACK,
        TRANSFORMATION,
        CURATION,
    ]

    def __init__(
        self,
        category: str,
        identifiers: List[Identifier],
        apply_fn: Callable = None,
        *args,
        **kwargs,
    ):

        super(SliceBuilder, self).__init__(*args, **kwargs)

        # The SliceMaker belongs to a category
        assert (
            category in self.CATEGORIES
        ), f"argument category must be one of {self.CATEGORIES}"
        self.category = category

        # Each identifier corresponds to a single output Slice generated by this
        # SliceBuilder
        self.identifiers = identifiers

        # Keep track of the CachedOperation dependencies
        self.prerequisites = (
            set() if "prerequisites" not in kwargs else kwargs["prerequisites"]
        )
        # TODO(karan): remove comment
        # for base in self.__class__.__bases__:
        #     for cls in base.__mro__:
        #         if str(CachedOperation.__name__) in str(cls):
        #             self.prerequisites.add(base)

        if apply_fn:
            # Assign to the method
            self.apply = apply_fn

    def __repr__(self):
        return (
            f"{self.category}[{self.__class__.__name__}(num_slices={self.num_slices})]"
        )

    @property
    def num_slices(self):
        return len(self.identifiers)

    def __getitem__(self, item: int):
        return self.identifiers[item]

    def __iter__(self):
        yield from self.identifiers

    def __call__(
        self,
        batch_or_dataset: BatchOrDataset,
        columns: List[str],
        mask: List[int] = None,
        num_proc: int = None,
        batch_size: int = 100,
        *args,
        **kwargs,
    ):

        # Check that prerequisites are satisfied
        self.prerequisites_handler(batch_or_dataset)

        if isinstance(batch_or_dataset, Dataset):
            # Prepare the dataset
            self.prepare_dataset(
                dataset=batch_or_dataset,
                columns=columns,
                batch_size=batch_size,
                *args,
                **kwargs,
            )

            # Slice a dataset
            slices, slice_membership = self.process_dataset(
                dataset=batch_or_dataset,
                columns=columns,
                # # Automatically infer the mask from the Dataset if it's not specified
                # mask=batch_or_dataset.check_tape(
                #     path=[SLICEBUILDERS, self.category],
                #     identifiers=self.identifiers,
                #     columns=columns,
                # )
                # if not mask
                # else mask,
                num_proc=num_proc,
                *args,
                **kwargs,
            )

            # Update the Dataset's history
            # TODO(karan): use mask to figure out what is actually applied
            batch_or_dataset.update_tape(
                path=[SLICEBUILDERS, self.category],
                identifiers=self.identifiers,
                columns=columns,
            )

            return slices, slice_membership

        elif isinstance(batch_or_dataset, Dict):
            # Slice a batch
            return self.process_batch(
                batch=batch_or_dataset,
                columns=columns,
                *args,
                **kwargs,
            )
        else:
            raise NotImplementedError

    def prerequisites_handler(
        self,
        batch_or_dataset: BatchOrDataset,
    ):
        if isinstance(batch_or_dataset, Dataset):
            batch = batch_or_dataset[:2]
        else:
            batch = batch_or_dataset

        # Check if prerequisites are satisfied
        # TODO(karan): move to a method
        pending = {
            prerequisite
            for prerequisite in self.prerequisites
            if not prerequisite.available(batch)
        }

        # TODO(karan): Automatically run the pending pre-requisites
        if pending:
            raise RuntimeError(
                f"Cannot run SliceBuilder, prerequisites {pending} not satisfied."
            )

    @staticmethod
    def store(batch: Dict[str, List], updates: List[Dict]) -> Dict[str, List]:
        """Update a batch of examples with slice information."""
        if "slices" not in batch:
            batch["slices"] = [{} for _ in range(len(batch["index"]))]

        # For each example, recursively merge the example's original cache dictionary
        # with the update dictionary
        batch["slices"] = [
            recmerge(example_dict, update_dict, merge_sequences=True)
            for example_dict, update_dict in zip(batch["slices"], updates)
        ]

        return batch

    def prepare_batch(
        self,
        batch: Batch,
        columns: List[str],
        *args,
        **kwargs,
    ) -> None:
        """Apply a preparation function to a batch. Use this to update
        attributes of `self`.

        Args:
            batch: batch of data
            columns: list of columns
            *args: optional additional arguments
            **kwargs: optional additional keyword arguments
        """
        raise NotImplementedError("Implemented `prepare_batch` should return a batch.")

    def _filter_prerequisite_columns(
        self,
        columns: List[str],
        all_columns: List[str],
    ) -> List[str]:
        # Simple filtering that doesn't use columns
        # TODO(karan): improve this by using `columns` to filter further
        return [
            col
            for col in all_columns
            if any(
                [
                    col.startswith(
                        prereq.__name__ if isinstance(prereq, type) else str(prereq)
                    )
                    for prereq in self.prerequisites
                ]
            )
        ]

    def prepare_dataset(
        self,
        dataset: Dataset,
        columns: List[str],
        batch_size: int = 32,
        *args,
        **kwargs,
    ) -> None:
        """Apply a preparation function to the dataset. Use this to update
        attributes of `self`.

        Args:
            dataset: dataset
            columns: list of columns
            batch_size: batch size for preparation
            *args: optional additional arguments
            **kwargs: optional additional keyword arguments
        """
        # Set the data format
        with dataset.format(
            columns + self._filter_prerequisite_columns(columns, dataset.column_names)
        ):
            # Batch the dataset, and prepare each batch
            for batch in dataset.batch(batch_size):
                try:
                    # Check if the `prepare_batch` function has been implemented
                    self.prepare_batch(
                        batch=batch,
                        columns=columns,
                        *args,
                        **kwargs,
                    )
                except NotImplementedError:
                    break

    def process_batch(
        self,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs,
    ) -> Tuple[List[Dict[str, List]], Optional[np.ndarray]]:
        """Apply a SliceBuilder to a batch of data.

        Args:
            batch: a batch of data
            columns: list of columns
            *args: optional additional arguments
            **kwargs: optional additional keyword arguments

        Returns: tuple of (list of slices (as batches), matrix of (example,
        slice) membership))
        """
        return [batch], None

    def process_dataset(
        self,
        dataset: Dataset,
        columns: List[str],
        batch_size: int = 32,
        # mask: List[int] = None,
        num_proc: int = None,
        *args,
        **kwargs,
    ) -> Tuple[List[Slice], np.ndarray]:
        """Apply a SliceBuilder to a dataset.

        Args:
            dataset: Dataset
            columns: list of columns
            batch_size: integer batch size
            # mask: boolean or integer mask array, mask[i] = True means that the ith
            # slice will be masked out
            num_proc: num processes for multiprocessing
            *args: optional additional arguments
            **kwargs: optional additional keyword arguments

        Returns: tuple of (Dataset, list of Slices, matrix of (example,
        slice) membership)
        """

        # # Compute a hash
        # val = persistent_hash(str(dataset.identifier)) ^ dataset.hash_interactions()
        # for i, identifier in enumerate(self.identifiers):
        #     if not mask[i]:
        #         val ^= persistent_hash(str(identifier)
        #         + str(strings_as_json(columns)))

        # try:
        #     # Map the SliceBuilder over the dataset
        #     all_sliced_batches = []
        #     all_slice_memberships = []
        #
        #     def _map_fn(batch):
        #         """Map function for processing batches.
        #
        #         Note that using this map_fn in a stateful way is
        #         dangerous, since every invocation of this function
        #         appends to the all_slice_batches list. The .map()
        #         function will invoke this once for testing before
        #         performing the map, so we discard the first entry
        #         inserted into all_sliced_batches.
        #         """
        #         batch, sliced_batches, slice_membership = self.process_batch(
        #             batch=batch,
        #             columns=columns,
        #             mask=mask,
        #             store_compressed=store_compressed,
        #             store=store,
        #             *args,
        #             **kwargs,
        #         )
        #         all_sliced_batches.append(sliced_batches)
        #         all_slice_memberships.append(slice_membership)
        #         return batch
        #
        #     dataset = dataset.map(
        #         _map_fn,
        #         batched=True,
        #         batch_size=batch_size,
        #         # FIXME(karan): enable this by adding logic for generating
        #         #  all_sliced_batches and all_slice_memberships
        #         #  when loading from cache file
        #         load_from_cache_file=False,
        #         # The cache file name is a XOR of the interaction history and the
        #         # current operation
        #         cache_file_name=str(
        #             dataset.logdir / ("cache-" + str(abs(val)) + ".arrow")
        #         ),
        #     )
        #
        #     # Remove the first entry (see _map_fn)
        #     all_sliced_batches = all_sliced_batches[1:]
        #     all_slice_memberships = all_slice_memberships[1:]
        #
        # except:  # noqa
        # all_batches, all_sliced_batches, all_slice_memberships = zip(
        #     *[
        #         self.process_batch(
        #             batch=batch,
        #             columns=columns,
        #             mask=mask,
        #             store_compressed=store_compressed,
        #             store=store,
        #             *args,
        #             **kwargs,
        #         )
        #         for batch in dataset.batch(batch_size)
        #     ]
        # )
        # # Update the dataset efficiently by reusing all_batches
        # dataset = dataset.map(
        #     lambda examples, indices: all_batches[indices[0] // batch_size],
        #     batched=True,
        #     batch_size=batch_size,
        #     with_indices=True,
        #     load_from_cache_file=False,
        #     # The cache file name is a XOR of the interaction history and the
        #     # current operation
        #     cache_file_name=str(
        #         dataset.logdir / ("cache-" + str(abs(val)) + ".arrow")
        #     ),
        # )

        # Create slices
        slices = [Slice() for _ in range(len(self.identifiers))]
        all_slice_memberships = []
        # Batch the dataset, and process each batch
        for batch in dataset.batch(batch_size):
            # Process the batch
            sliced_batches, slice_memberships = self.process_batch(
                batch=batch,
                columns=columns,
                *args,
                **kwargs,
            )

            # Incrementally build the slices
            for sl, sl_batch in zip(slices, sliced_batches):
                sl._dataset.append(sl_batch)

            # Keep track of the slice memberships
            all_slice_memberships.append(slice_memberships)

        # Create a single slice label matrix
        slice_membership = np.concatenate(all_slice_memberships, axis=0)

        # slice_cache_hashes = []
        # for identifier in self.identifiers:
        #     slice_cache_hashes.append(val ^ persistent_hash(str(identifier)))

        # if not num_proc or num_proc == 1:
        #     # Construct slices
        #     slices = []
        #     for i, slice_batches in enumerate(zip(*all_sliced_batches)):
        #         slices.append(
        #             create_slice(
        #                 (
        #                     dataset,
        #                     slice_membership,
        #                     slice_batches,
        #                     i,
        #                     batch_size,
        #                     slice_cache_hashes[i],
        #                 )
        #             )
        #         )
        # else:
        #     # Parallelized slice construction
        #     with Pool(num_proc) as pool:
        #         slices = pool.map(
        #             create_slice,
        #             [
        #                 (
        #                     dataset,
        #                     slice_membership,
        #                     slice_batches,
        #                     i,
        #                     batch_size,
        #                     slice_cache_hashes[i],
        #                 )
        #                 for i, slice_batches in enumerate(zip(*all_sliced_batches))
        #             ],
        #         )

        for i, sl in enumerate(slices):
            # Set the Slice category using the SliceBuilder's category
            sl.category = self.category

            # Append the the lineage
            sl.add_to_lineage(
                category=str(self.category.capitalize()),
                identifier=self.identifiers[i],
                columns=strings_as_json(columns),
            )

            # # Create the lineage
            # sl.lineage = [
            #     (str(Dataset.__name__), dataset.identifier),
            #     (
            #         str(self.category.capitalize()),
            #         self.identifiers[i],
            #         strings_as_json(columns),
            #     ),
            # ]
            # if isinstance(dataset, Slice):
            #     # Prepend the Slice's lineage instead, if the dataset was a slice
            #     sl.lineage = dataset.lineage + [
            #         (
            #             str(self.category.capitalize()),
            #             self.identifiers[i],
            #             strings_as_json(columns),
            #         )
            #     ]

        return slices, slice_membership

    def apply(self, *args, **kwargs):
        raise NotImplementedError("Must implement apply.")

    @classmethod
    def join(cls, *slicemakers: SliceBuilder) -> Sequence[SliceBuilder]:
        """Join many slicemakers.

        By default, just returns the slicemakers.
        """
        return slicemakers

    def masked(self, mask: List[int]):
        pass

    def unmasked(self):
        pass

    @staticmethod
    def filter_batch_by_slice_membership(
        batch: Dict[str, List], slice_membership: np.ndarray
    ) -> List[Dict[str, List]]:
        """Use a matrix of slice membership labels to select the subset of
        examples in each slice.

        Returns a list. Each element in the list corresponds to a single
        slice, and contains the subset of examples in 'batch' that lies
        in that slice.
        """
        return [
            tz.valmap(lambda v: list(compress(v, s)), batch) for s in slice_membership.T
        ]

    @classmethod
    def retrieve(
        cls,
        batch: Batch,
        columns: Union[List[str], List[List[str]]],
        proc_fns: Union[str, Callable, List[Union[str, Callable]]] = None,
        identifier: Union[str, Identifier] = None,
        reapply: bool = False,
        **kwargs,
    ) -> Optional[Union[Batch, List[Batch]]]:
        if not reapply:
            if "slices" not in batch:
                return None

            # Infer the most relevant key to retrieve if an identifier is not specified
            if not identifier:
                for ident_key in batch["slices"][0].keys():
                    # Pick the first key that matches the cls name
                    if ident_key.startswith(cls.__name__):
                        identifier = ident_key
                        break

            try:
                if isinstance(columns[0], str):
                    retrieval = {
                        strings_as_json(columns): [
                            cls.decode(cache[str(identifier)][strings_as_json(columns)])
                            for cache in batch["cache"]
                        ]
                    }
                else:
                    retrieval = {
                        strings_as_json(cols_): [
                            cls.decode(cache[str(identifier)][strings_as_json(cols_)])
                            for cache in batch["cache"]
                        ]
                        for cols_ in columns
                    }
            except KeyError:
                raise ValueError("Could not retrieve information for all keys.")

            # Check if the retrieved information needs to be processed
            if not proc_fns:
                return retrieval
            pass
        else:
            pass


class SliceBuilderCollection(SliceBuilder):
    """Collection of Slice Builders."""

    def __init__(self, slicebuilders: List[SliceBuilder], *args, **kwargs):
        super(SliceBuilderCollection, self).__init__(
            category=GENERIC,
            identifiers=list(
                tz.concat([slicebuilder.identifiers for slicebuilder in slicebuilders])
            ),
            *args,
            **kwargs,
        )

        # TODO(karan): some slicebuilders aren't compatible with each other (e.g.
        #  single column vs. multi column):
        #  add some smarter logic here to handle this

        # Store the subpopulations
        self.slicebuilders = slicebuilders

    def __repr__(self):
        # TODO(karan): format this nicely
        return (
            f"{self.__class__.__name__}("
            f"{[str(slicebuilder) for slicebuilder in self.slicebuilders]})]"
        )

    def __call__(
        self,
        batch_or_dataset: BatchOrDataset,
        columns: List[str],
        mask: List[int] = None,
        store_compressed: bool = None,
        store: bool = None,
        *args,
        **kwargs,
    ):

        if mask:
            raise NotImplementedError(
                "Mask not supported for SliceBuilderCollection yet."
            )

        slices = []
        slice_membership = []
        # Apply each slicebuilder in sequence
        for i, slicebuilder in tqdm.tqdm(enumerate(self.slicebuilders)):
            # Apply the slicebuilder
            batch_or_dataset, slices_i, slice_membership_i = slicebuilder(
                batch_or_dataset=batch_or_dataset,
                columns=columns,
                mask=mask,
                store_compressed=store_compressed,
                store=store,
                *args,
                **kwargs,
            )

            # Add in the slices and slice membership
            slices.extend(slices_i)
            slice_membership.append(slice_membership_i)

        slice_membership = np.concatenate(slice_membership, axis=1)

        return batch_or_dataset, slices, slice_membership


def create_slice(args):
    # Unpack args
    dataset, slice_membership, slice_batches, i, batch_size, slice_cache_hash = args

    # Create a new empty slice
    sl = Slice.from_dict({})

    # Create a Slice "copy" of the Dataset
    sl.__dict__.update(dataset.__dict__)
    sl._identifier = None

    # Filter
    sl = sl.filter(
        lambda example, idx: bool(slice_membership[idx, i]),
        with_indices=True,
        input_columns=["index"],
        batch_size=batch_size,
        cache_file_name=str(
            dataset.logdir / ("cache-" + str(abs(slice_cache_hash)) + "-filter.arrow")
        ),
    )

    slice_batch = tz.merge_with(tz.compose(list, tz.concat), slice_batches)

    # FIXME(karan): interaction tape history is wrong here, esp with augmenation/attacks

    # Map
    if len(sl):
        sl = sl.map(
            lambda batch, indices: tz.valmap(
                lambda v: v[indices[0] : indices[0] + batch_size], slice_batch
            ),
            batched=True,
            batch_size=batch_size,
            with_indices=True,
            remove_columns=sl.column_names,
            cache_file_name=str(
                dataset.logdir / ("cache-" + str(abs(slice_cache_hash)) + ".arrow")
            ),
        )

    return sl
